from brian2 import *
from brian2.tests.features import *
from brian2.tests.features.base import *
from brian2.tests.features.monitors import SpikeMonitorTest, StateMonitorTest
from brian2.tests.features.input import SpikeGeneratorGroupTest

# Full testing
print(run_feature_tests().tables_and_exceptions)

# Quick testing
# print run_feature_tests(configurations=[DefaultConfiguration,
#                                         NumpyConfiguration],
#                         feature_tests=[SpikeGeneratorGroupTest]).tables_and_exceptions

# Specific testing
#from brian2.tests.features.synapses import SynapsesSTDP, SynapsesPost
#print run_feature_tests(feature_tests=[SynapsesPost]).tables_and_exceptions
#print run_feature_tests(feature_tests=[SynapsesPost],
#                        configurations=[DefaultConfiguration]).exceptions



"""
Brian AST representation

This is a standard Python AST representation with additional information added.
"""

import ast
import weakref

import numpy

from brian2.utils.logger import get_logger

__all__ = ["brian_ast", "BrianASTRenderer", "dtype_hierarchy"]


logger = get_logger(__name__)

# This codifies the idea that operations involving e.g. boolean and integer will end up
# as integer. In general the output type will be the max of the hierarchy values here.
dtype_hierarchy = {
    "boolean": 0,
    "integer": 1,
    "float": 2,
}
# This is just so you can invert from number to string
for tc, i in dict(dtype_hierarchy).items():
    dtype_hierarchy[i] = tc


def is_boolean(value):
    return isinstance(value, bool)


def is_integer(value):
    return isinstance(value, (int, numpy.integer))


def is_float(value):
    return isinstance(value, (float, numpy.float32, numpy.float64))


def brian_dtype_from_value(value):
    """
    Returns 'boolean', 'integer' or 'float'
    """
    if is_float(value):
        return "float"
    elif is_integer(value):
        return "integer"
    elif is_boolean(value):
        return "boolean"
    raise TypeError(f"Unknown dtype for value {str(value)}")


# The following functions are called very often during the optimisation process
# so we don't use numpy.issubdtype but instead a precalculated list of all
# standard types

bool_dtype = numpy.dtype(bool)


def is_boolean_dtype(obj):
    return numpy.dtype(obj) is bool_dtype


integer_dtypes = {numpy.dtype(c) for c in numpy.typecodes["AllInteger"]}


def is_integer_dtype(obj):
    return numpy.dtype(obj) in integer_dtypes


float_dtypes = {numpy.dtype(c) for c in numpy.typecodes["AllFloat"]}


def is_float_dtype(obj):
    return numpy.dtype(obj) in float_dtypes


def brian_dtype_from_dtype(dtype):
    """
    Returns 'boolean', 'integer' or 'float'
    """
    if is_float_dtype(dtype):
        return "float"
    elif is_integer_dtype(dtype):
        return "integer"
    elif is_boolean_dtype(dtype):
        return "boolean"
    raise TypeError(f"Unknown dtype: {str(dtype)}")


def brian_ast(expr, variables):
    """
    Returns an AST tree representation with additional information

    Each node will be a standard Python ``ast`` node with the
    following additional attributes:

    ``dtype``
        One of ``'boolean'``, ``'integer'`` or ``'float'``, referring to the data type
        of the value of this node.
    ``scalar``
        Either ``True`` or ``False`` if the node uses any vector-valued variables.
    ``complexity``
        An integer representation of the computational complexity of the node. This
        is a very rough representation used for things like ``2*(x+y)`` is less
        complex than ``2*x+2*y`` and ``exp(x)`` is more complex than ``2*x`` but
        shouldn't be relied on for fine distinctions between expressions.

    Parameters
    ----------
    expr : str
        The expression to convert into an AST representation
    variables : dict
        The dictionary of `Variable` objects used in the expression.
    """
    node = ast.parse(expr, mode="eval").body
    renderer = BrianASTRenderer(variables)
    return renderer.render_node(node)


class BrianASTRenderer:
    """
    This class is modelled after `NodeRenderer` - see there for details.
    """

    def __init__(self, variables, copy_variables=True):
        if copy_variables:
            self.variables = variables.copy()
        else:
            self.variables = variables

    def render_node(self, node):
        nodename = node.__class__.__name__
        methname = f"render_{nodename}"
        try:
            return getattr(self, methname)(node)
        except AttributeError:
            raise SyntaxError(f"Unknown syntax: {nodename}")

    def render_NameConstant(self, node):
        if node.value is not True and node.value is not False:
            raise SyntaxError(f"Unknown NameConstant {str(node.value)}")
        # NameConstant only used for True and False and None, and we don't support None
        node.dtype = "boolean"
        node.scalar = True
        node.complexity = 0
        node.stateless = True
        return node

    def render_Name(self, node):
        node.complexity = 0
        if node.id == "True" or node.id == "False":
            node.dtype = "boolean"
            node.scalar = True
        elif node.id in self.variables:
            var = self.variables[node.id]
            dtype = var.dtype
            node.dtype = brian_dtype_from_dtype(dtype)
            node.scalar = var.scalar
        else:  # don't think we need to handle other names (pi, e, inf)?
            node.dtype = "float"
            node.scalar = True  # I think this assumption is OK, but not certain
        node.stateless = True
        return node

    def render_Num(self, node):
        node.complexity = 0
        node.dtype = brian_dtype_from_value(node.value)
        node.scalar = True
        node.stateless = True
        return node

    def render_Constant(self, node):  # For literals in Python >= 3.8
        if node.value is True or node.value is False or node.value is None:
            return self.render_NameConstant(node)
        else:
            return self.render_Num(node)

    def render_Call(self, node):
        if len(node.keywords):
            raise ValueError("Keyword arguments not supported.")
        elif getattr(node, "starargs", None) is not None:
            raise ValueError("Variable number of arguments not supported")
        elif getattr(node, "kwargs", None) is not None:
            raise ValueError("Keyword arguments not supported")
        args = []
        for subnode in node.args:
            subnode.parent = weakref.proxy(node)
            subnode = self.render_node(subnode)
            args.append(subnode)
        node.args = args
        node.dtype = "float"  # default dtype
        # Condition for scalarity of function call: stateless and arguments are scalar
        node.scalar = False
        if node.func.id in self.variables:
            funcvar = self.variables[node.func.id]
            # sometimes this attribute doesn't exist, if so assume it's not stateless
            node.stateless = getattr(funcvar, "stateless", False)
            node.auto_vectorise = getattr(funcvar, "auto_vectorise", False)
            if node.stateless and not node.auto_vectorise:
                node.scalar = all(subnode.scalar for subnode in node.args)
            # check that argument types are valid
            node_arg_types = [subnode.dtype for subnode in node.args]
            for subnode, argtype in zip(node.args, funcvar._arg_types):
                if argtype != "any" and argtype != subnode.dtype:
                    raise TypeError(
                        f"Function '{node.func.id}' takes arguments with "
                        f"types {funcvar._arg_types} but "
                        f"received {node_arg_types}."
                    )
            # compute return type
            return_type = funcvar._return_type
            if return_type == "highest":
                return_type = dtype_hierarchy[
                    max(dtype_hierarchy[nat] for nat in node_arg_types)
                ]
            node.dtype = return_type
        else:
            node.stateless = False
        # we leave node.func because it is an ast.Name object that doesn't have a dtype
        # TODO: variable complexity for function calls?
        node.complexity = 20 + sum(subnode.complexity for subnode in node.args)
        return node

    def render_BinOp(self, node):
        node.left.parent = weakref.proxy(node)
        node.right.parent = weakref.proxy(node)
        node.left = self.render_node(node.left)
        node.right = self.render_node(node.right)
        # TODO: we could capture some syntax errors here, e.g. bool+bool
        # captures, e.g. int+float->float
        newdtype = dtype_hierarchy[
            max(dtype_hierarchy[subnode.dtype] for subnode in [node.left, node.right])
        ]
        if node.op.__class__.__name__ == "Div":
            # Division turns integers into floating point values
            newdtype = "float"
        node.dtype = newdtype
        node.scalar = node.left.scalar and node.right.scalar
        node.complexity = 1 + node.left.complexity + node.right.complexity
        node.stateless = node.left.stateless and node.right.stateless
        return node

    def render_BoolOp(self, node):
        values = []
        for subnode in node.values:
            subnode.parent = node
            subnode = self.render_node(subnode)
            values.append(subnode)
        node.values = values
        node.dtype = "boolean"
        for subnode in node.values:
            if subnode.dtype != "boolean":
                raise TypeError("Boolean operator acting on non-booleans")
        node.scalar = all(subnode.scalar for subnode in node.values)
        node.complexity = 1 + sum(subnode.complexity for subnode in node.values)
        node.stateless = all(subnode.stateless for subnode in node.values)
        return node

    def render_Compare(self, node):
        node.left = self.render_node(node.left)
        comparators = []
        for subnode in node.comparators:
            subnode.parent = node
            subnode = self.render_node(subnode)
            comparators.append(subnode)
        node.comparators = comparators
        node.dtype = "boolean"
        comparators = [node.left] + node.comparators
        node.scalar = all(subnode.scalar for subnode in comparators)
        node.complexity = 1 + sum(subnode.complexity for subnode in comparators)
        node.stateless = node.left.stateless and all(
            c.stateless for c in node.comparators
        )
        return node

    def render_UnaryOp(self, node):
        node.operand.parent = node
        node.operand = self.render_node(node.operand)
        node.dtype = node.operand.dtype
        if node.dtype == "boolean" and node.op.__class__.__name__ != "Not":
            raise TypeError(
                f"Unary operator {node.op.__class__.__name__} does not apply to boolean"
                " types"
            )
        node.scalar = node.operand.scalar
        node.complexity = 1 + node.operand.complexity
        node.stateless = node.operand.stateless
        return node



"""
This example use C++ standalone mode for the simulation and the
`joblib library <https://joblib.readthedocs.io>`_
to parallelize the code. See the previous example (``02_using_standalone.py``)
for more explanations.
"""
from joblib import Parallel, delayed
from time import time as wall_time
from brian2 import *
import os


def run_sim(tau):
    pid = os.getpid()
    directory = f"standalone{pid}"
    set_device('cpp_standalone', directory=directory)
    print(f'RUNNING {pid}')

    G = NeuronGroup(1, 'dv/dt = -v/tau : 1', method='euler')
    G.v = 1

    mon = StateMonitor(G, 'v', record=0)
    net = Network()
    net.add(G, mon)
    net.run(100 * ms)
    res = (mon.t/ms, mon.v[0])

    device.reinit()

    print(f'FINISHED {pid}')
    return res


if __name__ == "__main__":
    start_time = wall_time()
    
    n_jobs = 4
    tau_values = np.arange(10)*ms + 5*ms

    results = Parallel(n_jobs=n_jobs)(map(delayed(run_sim), tau_values))

    print(f"Done in {wall_time() - start_time:10.3f}")

    for tau_value, (t, v) in zip(tau_values, results):
        plt.plot(t, v, label=str(tau_value))
    plt.legend()
    plt.show()



'''
This tool can be used to run setup.py inplace on Windows.
'''

import os
os.chdir('../../')
if os.path.exists('brian2/synapses/cythonspikequeue.pyd'):
    os.remove('brian2/synapses/cythonspikequeue.pyd')
if os.path.exists('brian2/synapses/cythonspikequeue.cpp'):
    os.remove('brian2/synapses/cythonspikequeue.cpp')
os.system('python setup.py build_ext --inplace')



import functools

import pytest

from brian2 import *
from brian2.codegen.runtime.GSLcython_rt import IntegrationError
from brian2.core.preferences import PreferenceError
from brian2.stateupdaters.base import UnsupportedEquationsException
from brian2.tests.utils import exc_isinstance

max_difference = 0.1 * mV

pytestmark = pytest.mark.gsl


def skip_if_not_implemented(func):
    @functools.wraps(func)
    def wrapped():
        try:
            func()
        except (BrianObjectException, NotImplementedError) as exc:
            if not (
                isinstance(exc, NotImplementedError)
                or isinstance(exc.__cause__, NotImplementedError)
            ):
                raise
            pytest.skip("GSL support for numpy has not been implemented yet")

    return wrapped


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_stateupdater_basic():
    # just the adaptive_threshold example: run for exponential_euler and GSL and see
    # if results are comparable (same amount of spikes and spikes > 0)
    eqs = """
    dv/dt = -v/(10*ms) : volt
    dvt/dt = (10*mV-vt)/(15*ms) : volt
    """
    reset = """
    v = 0*mV
    vt += 3*mV
    """
    neurons_conventional = NeuronGroup(
        1, model=eqs, reset=reset, threshold="v>vt", method="exponential_euler"
    )
    neurons_GSL = NeuronGroup(1, model=eqs, reset=reset, threshold="v>vt", method="gsl")
    neurons_conventional.vt = 10 * mV
    neurons_GSL.vt = 10 * mV
    # 50 'different' neurons so no neuron spikes more than once per dt
    P = SpikeGeneratorGroup(1, [0] * 50, np.arange(50) / 50.0 * 100 * ms)
    C_conventional = Synapses(P, neurons_conventional, on_pre="v += 3*mV")
    C_GSL = Synapses(P, neurons_GSL, on_pre="v += 3*mV")
    C_conventional.connect()
    C_GSL.connect()
    SM_conventional = SpikeMonitor(neurons_conventional, variables="v")
    SM_GSL = SpikeMonitor(neurons_GSL, variables="v")
    net = Network(
        neurons_conventional,
        neurons_GSL,
        P,
        C_conventional,
        C_GSL,
        SM_conventional,
        SM_GSL,
    )
    net.run(100 * ms)
    assert (
        SM_conventional.num_spikes > 0
    ), "simulation should produce spiking, but no spikes monitored"
    assert SM_conventional.num_spikes == SM_GSL.num_spikes, (
        "GSL_statupdater produced different number of spikes than integration with ",
        "exponential euler",
    )


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_different_clocks():
    vt = 10 * mV
    eqs = "dv/dt = -v/(10*ms) : volt"
    neurons = NeuronGroup(1, model=eqs, threshold="v>vt", method="gsl", dt=0.2 * ms)
    # for this test just check if it compiles
    run(0 * ms)


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_default_function():
    # phase_locking example
    tau = 20 * ms
    n = 100
    b = 1.2  # constant current mean, the modulation varies
    freq = 10 * Hz
    eqs = """
        dv/dt = (-v + a * sin(2 * pi * freq * t) + b) / tau : 1
        a : 1
    """
    vrand = rand(n)
    neurons_conventional = NeuronGroup(
        n, model=eqs, threshold="v > 1", reset="v = 0", method="exponential_euler"
    )
    neurons_GSL = NeuronGroup(
        n, model=eqs, threshold="v > 1", reset="v = 0", method="gsl"
    )
    neurons_conventional.v = vrand
    neurons_GSL.v = vrand
    neurons_conventional.a = "0.05 + 0.7*i/n"
    neurons_GSL.a = "0.05 + 0.7*i/n"

    trace_conventional = StateMonitor(neurons_conventional, "v", record=50)
    trace_GSL = StateMonitor(neurons_GSL, "v", record=50)
    net = Network(neurons_conventional, neurons_GSL, trace_conventional, trace_GSL)
    net.run(10 * ms)

    assert (
        max(trace_conventional.v[0] - trace_GSL.v[0]) < max_difference / mV
    ), "difference between conventional and GSL output is larger than max_difference"


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_user_defined_function():
    # phase_locking example with user_defined sin
    eqs = """
    dv/dt = (-v + a * sin(2 * pi * freq * t) + b) / tau : 1
    a : 1
    """

    @implementation(
        "cpp",
        """
        double user_sin(double phase)
        {
            return sin(phase);
        }
        """,
    )
    @implementation(
        "cython",
        """
        cdef double user_sin(double phase):
            return sin(phase)
        """,
    )
    @check_units(phase=1, result=1)
    def user_sin(phase):
        raise Exception

    tau = 20 * ms
    n = 100
    b = 1.2  # constant current mean, the modulation varies
    freq = 10 * Hz
    eqs = """
    dv/dt = (-v + a * user_sin(2 * pi * freq * t) + b) / tau : 1
    a : 1
    """
    vrand = rand(n)
    neurons_conventional = NeuronGroup(
        n, model=eqs, threshold="v > 1", reset="v = 0", method="exponential_euler"
    )
    neurons_GSL = NeuronGroup(
        n, model=eqs, threshold="v > 1", reset="v = 0", method="gsl"
    )
    neurons_conventional.v = vrand
    neurons_GSL.v = vrand
    neurons_conventional.a = "0.05 + 0.7*i/n"
    neurons_GSL.a = "0.05 + 0.7*i/n"

    trace_conventional = StateMonitor(neurons_conventional, "v", record=50)
    trace_GSL = StateMonitor(neurons_GSL, "v", record=50)
    net = Network(neurons_conventional, neurons_GSL, trace_conventional, trace_GSL)
    net.run(10 * ms)

    assert (
        max(trace_conventional.v[0] - trace_GSL.v[0]) < max_difference / mV
    ), "difference between conventional and GSL output is larger than max_difference"
    # assert not all(trace_conventional.v[0]==trace_GSL.v[0]), \
    #         ('output of GSL stateupdater is exactly the same as Brians stateupdater (unlikely to be right)')


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_x_variable():
    neurons = NeuronGroup(
        2, "dx/dt = 300*Hz : 1", threshold="x>1", reset="x=0", method="gsl"
    )
    # just testing compilation
    run(0 * ms)


@pytest.mark.codegen_independent
def test_GSL_failing_directory():
    def set_dir(arg):
        prefs.GSL.directory = arg

    with pytest.raises(PreferenceError):
        set_dir(1)
    with pytest.raises(PreferenceError):
        set_dir("/usr/")
    with pytest.raises(PreferenceError):
        set_dir("/usr/blablabla/")


@pytest.mark.codegen_independent
@skip_if_not_implemented
def test_GSL_stochastic():
    tau = 20 * ms
    sigma = 0.015
    eqs = """
    dx/dt = (1.1 - x) / tau + sigma * (2 / tau)**.5 * xi : 1
    """
    neuron = NeuronGroup(1, eqs, method="gsl")
    net = Network(neuron)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms, namespace={"tau": tau, "sigma": sigma})
    assert exc_isinstance(
        exc, UnsupportedEquationsException, raise_not_implemented=True
    )


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_error_dimension_mismatch_unit():
    eqs = """
    dv/dt = (v0 - v)/(10*ms) : volt
    v0 : volt
    """
    options = {"absolute_error_per_variable": {"v": 1 * nS}}
    neuron = NeuronGroup(
        1,
        eqs,
        threshold="v > 10*mV",
        reset="v = 0*mV",
        method="gsl",
        method_options=options,
    )
    net = Network(neuron)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms, namespace={})
    assert exc_isinstance(exc, DimensionMismatchError, raise_not_implemented=True)


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_error_dimension_mismatch_dimensionless1():
    eqs = """
    dv/dt = (v0 - v)/(10*ms) : 1
    v0 : 1
    """
    options = {"absolute_error_per_variable": {"v": 1 * mV}}
    neuron = NeuronGroup(
        1, eqs, threshold="v > 10", reset="v = 0", method="gsl", method_options=options
    )
    net = Network(neuron)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms, namespace={})
    assert exc_isinstance(exc, DimensionMismatchError, raise_not_implemented=True)


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_error_dimension_mismatch_dimensionless2():
    eqs = """
    dv/dt = (v0 - v)/(10*ms) : volt
    v0 : volt
    """
    options = {"absolute_error_per_variable": {"v": 1e-3}}
    neuron = NeuronGroup(
        1,
        eqs,
        threshold="v > 10*mV",
        reset="v = 0*mV",
        method="gsl",
        method_options=options,
    )
    net = Network(neuron)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms, namespace={})
    assert exc_isinstance(exc, DimensionMismatchError, raise_not_implemented=True)


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_error_nonexisting_variable():
    eqs = """
    dv/dt = (v0 - v)/(10*ms) : volt
    v0 : volt
    """
    options = {"absolute_error_per_variable": {"dummy": 1e-3 * mV}}
    neuron = NeuronGroup(
        1,
        eqs,
        threshold="v > 10*mV",
        reset="v = 0*mV",
        method="gsl",
        method_options=options,
    )
    net = Network(neuron)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms, namespace={})
    assert exc_isinstance(exc, KeyError, raise_not_implemented=True)


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_error_incorrect_error_format():
    eqs = """
    dv/dt = (v0 - v)/(10*ms) : volt
    v0 : volt
    """
    options = {"absolute_error_per_variable": object()}
    neuron = NeuronGroup(
        1,
        eqs,
        threshold="v > 10*mV",
        reset="v = 0*mV",
        method="gsl",
        method_options=options,
    )
    net = Network(neuron)
    options2 = {"absolute_error": "not a float"}
    neuron2 = NeuronGroup(
        1,
        eqs,
        threshold="v > 10*mV",
        reset="v = 0*mV",
        method="gsl",
        method_options=options2,
    )
    net2 = Network(neuron2)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms, namespace={})
    assert exc_isinstance(exc, TypeError, raise_not_implemented=True)
    with pytest.raises(BrianObjectException) as exc:
        net2.run(0 * ms, namespace={})
    assert exc_isinstance(exc, TypeError, raise_not_implemented=True)


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_error_nonODE_variable():
    eqs = """
    dv/dt = (v0 - v)/(10*ms) : volt
    v0 : volt
    """
    options = {"absolute_error_per_variable": {"v0": 1e-3 * mV}}
    neuron = NeuronGroup(
        1,
        eqs,
        threshold="v > 10*mV",
        reset="v = 0*mV",
        method="gsl",
        method_options=options,
    )
    net = Network(neuron)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms, namespace={})
    assert exc_isinstance(exc, KeyError, raise_not_implemented=True)


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_error_bounds():
    runtime = 50 * ms
    error1 = 1e-2 * volt
    error2 = 1e-4 * volt
    error3 = 1e-6 * volt  # default error
    eqs = """
    dv/dt = (stimulus(t) + -v)/(.1*ms) : volt
    """
    stimulus = TimedArray(rand(int(runtime / (10 * ms))) * 3 * volt, dt=5 * ms)
    neuron1 = NeuronGroup(
        1,
        model=eqs,
        reset="v=0*mV",
        threshold="v>10*volt",
        method="gsl",
        method_options={"absolute_error_per_variable": {"v": error1}},
        dt=1 * ms,
    )
    neuron2 = NeuronGroup(
        1,
        model=eqs,
        reset="v=0*mV",
        threshold="v>10*volt",
        method="gsl",
        method_options={"absolute_error_per_variable": {"v": error2}},
        dt=1 * ms,
    )
    neuron3 = NeuronGroup(
        1,
        model=eqs,
        reset="v=0*mV",
        threshold="v>10*volt",
        method="gsl",
        method_options={"absolute_error_per_variable": {}},
        dt=1 * ms,
    )  # Uses default error
    neuron_control = NeuronGroup(1, model=eqs, method="linear", dt=1 * ms)
    mon1 = StateMonitor(neuron1, "v", record=True)
    mon2 = StateMonitor(neuron2, "v", record=True)
    mon3 = StateMonitor(neuron3, "v", record=True)
    mon_control = StateMonitor(neuron_control, "v", record=True)
    run(runtime)
    err1 = abs(mon1.v[0] - mon_control.v[0])
    err2 = abs(mon2.v[0] - mon_control.v[0])
    err3 = abs(mon3.v[0] - mon_control.v[0])
    assert (
        max(err1) < error1
    ), f"Error bound exceeded, error bound: {error1:e}, obtained error: {max(err1):e}"
    assert max(err2) < error2, "Error bound exceeded"
    assert max(err3) < error3, "Error bound exceeded"
    assert max(err1) > max(
        err2
    ), "The simulation with smaller error bound produced a bigger maximum error"
    assert max(err2) > max(
        err3
    ), "The simulation with smaller error bound produced a bigger maximum error"


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_non_autonomous():
    eqs = """dv/dt = sin(2*pi*freq*t)/ms : 1
             freq : Hz"""
    neuron = NeuronGroup(10, eqs, method="gsl")
    neuron.freq = "i*10*Hz + 10*Hz"
    neuron2 = NeuronGroup(10, eqs, method="euler")
    neuron2.freq = "i*10*Hz + 10*Hz"
    mon = StateMonitor(neuron, "v", record=True)
    mon2 = StateMonitor(neuron2, "v", record=True)
    run(20 * ms)
    abs_err = np.abs(mon.v.T - mon2.v.T)
    max_allowed = 1000 * np.finfo(prefs.core.default_float_dtype).eps
    assert np.max(abs_err) < max_allowed


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_non_autonomous():
    eqs = """dv/dt = sin(2*pi*freq*t)/ms : 1
             freq : Hz"""
    neuron = NeuronGroup(10, eqs, method="gsl")
    neuron.freq = "i*10*Hz + 10*Hz"
    neuron2 = NeuronGroup(10, eqs, method="euler")
    neuron2.freq = "i*10*Hz + 10*Hz"
    mon = StateMonitor(neuron, "v", record=True)
    mon2 = StateMonitor(neuron2, "v", record=True)
    run(20 * ms)
    abs_err = np.abs(mon.v.T - mon2.v.T)
    max_allowed = 1000 * np.finfo(prefs.core.default_float_dtype).eps
    assert np.max(abs_err) < max_allowed


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_refractory():
    eqs = """dv/dt = 99.99*Hz : 1 (unless refractory)"""
    neuron = NeuronGroup(
        1, eqs, method="gsl", threshold="v>1", reset="v=0", refractory=3 * ms
    )
    neuron2 = NeuronGroup(
        1, eqs, method="euler", threshold="v>1", reset="v=0", refractory=3 * ms
    )
    mon = SpikeMonitor(neuron, "v")
    mon2 = SpikeMonitor(neuron2, "v")
    run(20 * ms)
    assert mon.count[0] == mon2.count[0]


@skip_if_not_implemented
def test_GSL_save_step_count():
    eqs = """
    dv/dt = -v/(.1*ms) : volt
    """
    neuron = NeuronGroup(
        1, model=eqs, method="gsl", method_options={"save_step_count": True}, dt=1 * ms
    )
    run(0 * ms)
    mon = StateMonitor(neuron, "_step_count", record=True, when="end")
    run(10 * ms)
    assert mon._step_count[0][0] > 0, "Monitor did not save GSL step count"


HH_namespace = {
    "Cm": 1 * ufarad * cm**-2,
    "gl": 5e-5 * siemens * cm**-2,
    "El": -65 * mV,
    "EK": -90 * mV,
    "ENa": 50 * mV,
    "g_na": 100 * msiemens * cm**-2,
    "g_kd": 30 * msiemens * cm**-2,
    "VT": -63 * mV,
}

HH_eqs = Equations(
    """
    dv/dt = (gl*(El-v) - g_na*(m*m*m)*h*(v-ENa) - g_kd*(n*n*n*n)*(v-EK) + I)/Cm : volt
    dm/dt = 0.32*(mV**-1)*(13.*mV-v+VT)/
        (exp((13.*mV-v+VT)/(4.*mV))-1.)/ms*(1-m)-0.28*(mV**-1)*(v-VT-40.*mV)/
        (exp((v-VT-40.*mV)/(5.*mV))-1.)/ms*m : 1
    dn/dt = 0.032*(mV**-1)*(15.*mV-v+VT)/
        (exp((15.*mV-v+VT)/(5.*mV))-1.)/ms*(1.-n)-.5*exp((10.*mV-v+VT)/(40.*mV))/ms*n : 1
    dh/dt = 0.128*exp((17.*mV-v+VT)/(18.*mV))/ms*(1.-h)-4./(1+exp((40.*mV-v+VT)/(5.*mV)))/ms*h : 1
    I : amp/metre**2
    """
)


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_fixed_timestep_big_dt_small_error():
    # should raise integration error
    neuron = NeuronGroup(
        1,
        model=HH_eqs,
        threshold="v > -40*mV",
        refractory="v > -40*mV",
        method="gsl",
        method_options={"adaptable_timestep": False, "absolute_error": 1e-12},
        dt=0.001 * ms,
        namespace=HH_namespace,
    )
    neuron.I = 0.7 * nA / (20000 * umetre**2)
    neuron.v = HH_namespace["El"]
    net = Network(neuron)
    with pytest.raises((BrianObjectException, RuntimeError)):
        net.run(10 * ms)


@pytest.mark.codegen_independent
@skip_if_not_implemented
def test_GSL_internal_variable():
    with pytest.raises(SyntaxError):
        Equations("d_p/dt = 300*Hz : 1")


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_method_options_neurongroup():
    neuron1 = NeuronGroup(
        1,
        model="dp/dt = 300*Hz : 1",
        method="gsl",
        method_options={"adaptable_timestep": True},
    )
    neuron2 = NeuronGroup(
        1,
        model="dp/dt = 300*Hz : 1",
        method="gsl",
        method_options={"adaptable_timestep": False},
    )
    run(0 * ms)
    assert "if (gsl_odeiv2_driver_apply_fixed_step" not in str(
        neuron1.state_updater.codeobj.code
    ), "This neuron should not call gsl_odeiv2_driver_apply_fixed_step()"
    assert "if (gsl_odeiv2_driver_apply_fixed_step" in str(
        neuron2.state_updater.codeobj.code
    ), "This neuron should call gsl_odeiv2_driver_apply_fixed_step()"


@pytest.mark.standalone_compatible
@skip_if_not_implemented
def test_GSL_method_options_spatialneuron():
    morpho = Soma(30 * um)
    eqs = """
    Im = g * v : amp/meter**2
    dg/dt = siemens/metre**2/second : siemens/metre**2
    """
    neuron1 = SpatialNeuron(
        morphology=morpho,
        model=eqs,
        Cm=1 * uF / cm**2,
        Ri=100 * ohm * cm,
        method="gsl_rkf45",
        method_options={"adaptable_timestep": True},
    )
    neuron2 = SpatialNeuron(
        morphology=morpho,
        model=eqs,
        Cm=1 * uF / cm**2,
        Ri=100 * ohm * cm,
        method="gsl_rkf45",
        method_options={"adaptable_timestep": False},
    )
    run(0 * ms)
    assert "if (gsl_odeiv2_driver_apply_fixed_step" not in str(
        neuron1.state_updater.codeobj.code
    ), "This neuron should not call gsl_odeiv2_driver_apply_fixed_step()"
    assert "if (gsl_odeiv2_driver_apply_fixed_step" in str(
        neuron2.state_updater.codeobj.code
    ), "This neuron should call gsl_odeiv2_driver_apply_fixed_step()"


@skip_if_not_implemented
def test_GSL_method_options_synapses():
    N = 1000
    taum = 10 * ms
    taupre = 20 * ms
    taupost = taupre
    Ee = 0 * mV
    vt = -54 * mV
    vr = -60 * mV
    El = -74 * mV
    taue = 5 * ms
    F = 15 * Hz
    gmax = 0.01
    dApre = 0.01
    dApost = -dApre * taupre / taupost * 1.05
    dApost *= gmax
    dApre *= gmax
    eqs_neurons = """
    dv/dt = (ge * (Ee-vr) + El - v) / taum : volt
    dge/dt = -ge / taue : 1
    """
    poisson_input = PoissonGroup(N, rates=F)
    neurons = NeuronGroup(
        1, eqs_neurons, threshold="v>vt", reset="v = vr", method="gsl_rkf45"
    )
    S1 = Synapses(
        poisson_input,
        neurons,
        """
        w : 1
        dApre/dt = -Apre / taupre : 1 (clock-driven)
        dApost/dt = -Apost / taupost : 1 (clock-driven)
        """,
        method="gsl_rkf45",
        method_options={"adaptable_timestep": True},
    )
    S2 = Synapses(
        poisson_input,
        neurons,
        """
        w : 1
        dApre/dt = -Apre / taupre : 1 (clock-driven)
        dApost/dt = -Apost / taupost : 1 (clock-driven)
        """,
        method="gsl_rkf45",
        method_options={"adaptable_timestep": False},
    )
    run(0 * ms)
    assert "if (gsl_odeiv2_driver_apply_fixed_step" not in str(
        S1.state_updater.codeobj.code
    ), "This state_updater should not call gsl_odeiv2_driver_apply_fixed_step()"
    assert "if (gsl_odeiv2_driver_apply_fixed_step" in str(
        S2.state_updater.codeobj.code
    ), "This state_updater should call gsl_odeiv2_driver_apply_fixed_step()"


if __name__ == "__main__":
    from _pytest.outcomes import Skipped

    for t in [
        test_GSL_stateupdater_basic,
        test_GSL_different_clocks,
        test_GSL_default_function,
        test_GSL_user_defined_function,
        test_GSL_x_variable,
        test_GSL_failing_directory,
        test_GSL_stochastic,
        test_GSL_error_dimension_mismatch_unit,
        test_GSL_error_dimension_mismatch_dimensionless1,
        test_GSL_error_dimension_mismatch_dimensionless2,
        test_GSL_error_nonexisting_variable,
        test_GSL_error_incorrect_error_format,
        test_GSL_error_nonODE_variable,
        test_GSL_error_bounds,
        test_GSL_non_autonomous,
        test_GSL_refractory,
        test_GSL_save_step_count,
        test_GSL_fixed_timestep_big_dt_small_error,
        test_GSL_method_options_neurongroup,
        test_GSL_method_options_spatialneuron,
        test_GSL_method_options_synapses,
    ]:
        try:
            t()
        except Skipped as ex:
            print(f"Skipped: {t.__name__} ({str(ex)})")



import numpy as np
from numpy.testing import assert_allclose as numpy_allclose

from brian2 import prefs
from brian2.units.fundamentalunits import have_same_dimensions


def assert_allclose(actual, desired, rtol=4.5e8, atol=1e-9, **kwds):
    """
    Thin wrapper around numpy's `~numpy.testing.utils.assert_allclose` function. The tolerance depends on the floating
    point precision as defined by the `core.default_float_dtype` preference.

    Parameters
    ----------
    actual : `numpy.ndarray`
        The results to check.
    desired : `numpy.ndarray`
        The expected results.
    rtol : float, optional
        The relative tolerance which will be multiplied with the machine epsilon of the type set as
        `core.default_float_type`.
    atol : float, optional
        The absolute tolerance
    """
    assert have_same_dimensions(actual, desired)
    eps = np.finfo(prefs["core.default_float_dtype"]).eps
    rtol = eps * rtol
    numpy_allclose(
        np.asarray(actual), np.asarray(desired), rtol=rtol, atol=atol, **kwds
    )


def exc_isinstance(exc_info, expected_exception, raise_not_implemented=False):
    """
    Simple helper function as an alternative to calling
    `~.pytest.ExceptionInfo.errisinstance` which will take into account all
    the "causing" exceptions in an exception chain.

    Parameters
    ----------
    exc_info : `pytest.ExceptionInfo` or `Exception`
        The exception info as returned by `pytest.raises`.
    expected_exception : `type`
        The expected exception class
    raise_not_implemented : bool, optional
        Whether to re-raise a `NotImplementedError` – necessary for tests that
        should be skipped with ``@skip_if_not_implemented``. Defaults to
        ``False``.

    Returns
    -------
    correct_exception : bool
        Whether the exception itself or one of the causing exceptions is of the
        expected type.
    """
    if exc_info is None:
        return False
    if hasattr(exc_info, "value"):
        exc_info = exc_info.value

    if isinstance(exc_info, expected_exception):
        return True
    elif raise_not_implemented and isinstance(exc_info, NotImplementedError):
        raise exc_info

    return exc_isinstance(
        exc_info.__cause__,
        expected_exception,
        raise_not_implemented=raise_not_implemented,
    )



"""
Package contain all unit/integration tests for the `brian2` package.
"""

import os
import sys
import tempfile
from io import StringIO

import numpy as np

import brian2
from brian2.core.preferences import prefs
from brian2.devices.device import all_devices, reinit_and_delete, reset_device

try:
    import importlib

    import pytest
    from _pytest import doctest as pytest_doctest

    class OurDoctestModule(pytest_doctest.DoctestModule):
        def collect(self):
            for item in super().collect():
                # Check the object for exclusion from doctests
                full_name = item.name.split(".")
                test_name = []
                module_name = os.path.splitext(os.path.basename(self.name))[0]
                while full_name[-1] != module_name:
                    test_name.append(full_name.pop())
                tested_obj = self.obj
                for name in reversed(test_name):
                    tested_obj = getattr(tested_obj, name)
                if not getattr(tested_obj, "_do_not_run_doctests", False):
                    yield item

    # Monkey patch pytest
    pytest_doctest.DoctestModule = OurDoctestModule

except ImportError:
    pytest = None


class PreferencePlugin:
    def __init__(self, prefs, fail_for_not_implemented=True):
        self.prefs = prefs
        self.device = "runtime"
        self.device_options = {}
        self.fail_for_not_implemented = fail_for_not_implemented

    def pytest_configure(self, config):
        config.brian_prefs = dict(self.prefs)
        config.fail_for_not_implemented = self.fail_for_not_implemented
        config.device = self.device
        config.device_options = self.device_options
        if config.pluginmanager.hasplugin("xdist"):
            xdist_plugin = XDistPreferencePlugin(self)
            config.pluginmanager.register(xdist_plugin)


class XDistPreferencePlugin:
    def __init__(self, pref_plugin):
        self._pref_plugin = pref_plugin

    def pytest_configure_node(self, node):
        """xdist hook"""
        prefs = dict(self._pref_plugin.prefs)
        for k, v in prefs.items():
            if isinstance(v, type):
                prefs[k] = ("TYPE", repr(v))
        node.workerinput["brian_prefs"] = prefs
        node.workerinput["fail_for_not_implemented"] = (
            self._pref_plugin.fail_for_not_implemented
        )
        node.workerinput["device"] = self._pref_plugin.device
        node.workerinput["device_options"] = self._pref_plugin.device_options


def clear_caches():
    from brian2.utils.logger import BrianLogger

    BrianLogger._log_messages.clear()
    from brian2.codegen.translation import make_statements

    make_statements._cache.clear()


def make_argv(dirnames, markers=None, doctests=False, test_GSL=False):
    """
    Create the list of arguments for the ``pytests`` call.

    Parameters
    ----------
    markers : str, optional
        The markers of the tests to include.
    doctests : bool, optional
        Whether to run doctests. Defaults to ``False``.
    test_GSL : bool, optional
        Whether to run tests requiring the GSL. If set to
        ``False``, tests marked with ``gsl`` will be
        excluded. Defaults to ``False``.

    Returns
    -------
    argv : list of str
        The arguments for `pytest.main`.

    """
    if doctests:
        if markers is not None:
            raise TypeError("Cannot give markers for doctests")
        argv = dirnames + [
            "-c",
            os.path.join(os.path.dirname(__file__), "pytest.ini"),
            "--quiet",
            "--doctest-modules",
            "--doctest-glob=*.rst",
            "--doctest-ignore-import-errors",
            "--confcutdir",
            os.path.abspath(os.path.join(os.path.dirname(__file__), "..")),
            "--pyargs",
            "brian2",
        ]
        if len(dirnames) == 2:
            # If we are testing files in docs_sphinx, ignore conf.py
            argv += [f"--ignore={os.path.join(dirnames[1], 'conf.py')}"]
    else:
        if not test_GSL:
            markers += " and not gsl"
        argv = dirnames + [
            "-c",
            os.path.join(os.path.dirname(__file__), "pytest.ini"),
            "--quiet",
            "-m",
            f"{markers}",
            "--confcutdir",
            os.path.abspath(os.path.join(os.path.dirname(__file__), "..")),
        ]
    return argv


def run(
    codegen_targets=None,
    long_tests=False,
    test_codegen_independent=True,
    test_standalone=None,
    test_openmp=False,
    test_in_parallel=["codegen_independent", "numpy", "cython", "cpp_standalone"],
    reset_preferences=True,
    fail_for_not_implemented=True,
    test_GSL=False,
    build_options=None,
    extra_test_dirs=None,
    sphinx_dir=None,
    float_dtype=None,
    additional_args=None,
):
    """
    Run brian's test suite. Needs an installation of the pytest testing tool.

    For testing, the preferences will be reset to the default preferences.
    After testing, the user preferences will be restored.

    Parameters
    ----------
    codegen_targets : list of str or str
        A list of codegeneration targets or a single target, e.g.
        ``['numpy', 'cython']`` to test. The whole test suite will be repeatedly
        run with `codegen.target` set to the respective value. If not
        specified, all available code generation targets will be tested.
    long_tests : bool, optional
        Whether to run tests that take a long time. Defaults to ``False``.
    test_codegen_independent : bool, optional
        Whether to run tests that are independent of code generation. Defaults
        to ``True``.
    test_standalone : str, optional
        Whether to run tests for a standalone mode. Should be the name of a
        standalone mode (e.g. ``'cpp_standalone'``) and expects that a device
        of that name and an accordingly named "simple" device (e.g.
        ``'cpp_standalone_simple'`` exists that can be used for testing (see
        `CPPStandaloneSimpleDevice` for details. Defaults to ``None``, meaning
        that no standalone device is tested.
    test_openmp : bool, optional
        Whether to test standalone test with multiple threads and OpenMP. Will
        be ignored if ``cpp_standalone`` is not tested. Defaults to ``False``.
    reset_preferences : bool, optional
        Whether to reset all preferences to the default preferences before
        running the test suite. Defaults to ``True`` to get test results
        independent of the user's preference settings but can be switched off
        when the preferences are actually necessary to pass the tests (e.g. for
        device-specific settings).
    fail_for_not_implemented : bool, optional
        Whether to fail for tests raising a `NotImplementedError`. Defaults to
        ``True``, but can be switched off for devices known to not implement
        all of Brian's features.
    test_GSL : bool, optional
        Whether to test support for GSL state updaters (requires an installation
        of the GSL development packages). Defaults to ``False``.
    build_options : dict, optional
        Non-default build options that will be passed as arguments to the
        `set_device` call for the device specified in ``test_standalone``.
    extra_test_dirs : list of str or str, optional
        Additional directories as a list of strings (or a single directory as
        a string) that will be searched for additional tests.
    sphinx_dir : str, optional
        The full path to ``docs_sphinx``, in order to execute doc tests in the
        rst files. If not provided, assumes we are running from a checked out
        repository where the directory can be found at ``../../docs_sphinx``.
        Will ignore the provided directory if it does not exist.
    float_dtype : np.dtype, optional
        Set the dtype to use for floating point variables to a value different
        from the default `core.default_float_dtype` setting.
    additional_args : list of str, optional
        Optional command line arguments to pass to ``pytest``
    """
    if pytest is None:
        raise ImportError("Running the test suite requires the 'pytest' package.")

    if build_options is None:
        build_options = {}

    if os.name == "nt":
        test_in_parallel = []

    if extra_test_dirs is None:
        extra_test_dirs = []
    elif isinstance(extra_test_dirs, str):
        extra_test_dirs = [extra_test_dirs]
    if additional_args is None:
        additional_args = []

    multiprocess_arguments = ["-n", "auto"]

    if codegen_targets is None:
        codegen_targets = ["numpy"]
        try:
            import Cython

            codegen_targets.append("cython")
        except ImportError:
            pass
    elif isinstance(codegen_targets, str):  # allow to give a single target
        codegen_targets = [codegen_targets]

    dirname = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
    dirnames = [dirname] + extra_test_dirs

    print(f"Running tests in {', '.join(dirnames)} ", end="")
    if codegen_targets:
        print(f"for targets {', '.join(codegen_targets)}", end="")
    ex_in = "including" if long_tests else "excluding"
    print(f" ({ex_in} long tests)")

    print(
        f"Running Brian version {brian2.__version__} from"
        f" '{os.path.dirname(brian2.__file__)}'"
    )

    all_targets = set(codegen_targets)

    if test_standalone:
        if not isinstance(test_standalone, str):
            raise ValueError(
                "test_standalone argument has to be the name of a "
                "standalone device (e.g. 'cpp_standalone')"
            )
        if test_standalone not in all_devices:
            known_devices = ", ".join(repr(d) for d in all_devices)
            raise ValueError(
                "test_standalone argument 'test_standalone' is not a known "
                f"device. Known devices are: {known_devices}."
            )
        print("Testing standalone")
        all_targets.add(test_standalone)
    if test_codegen_independent:
        print("Testing codegen-independent code")
        all_targets.add("codegen_independent")

    parallel_tests = all_targets.intersection(set(test_in_parallel))
    if parallel_tests:
        try:
            import xdist

            print(f"Testing with multiple processes for {', '.join(parallel_tests)}")
        except ImportError:
            test_in_parallel = []

    if reset_preferences:
        print("Resetting to default preferences")
        stored_prefs = prefs.as_file
        prefs.reset_to_defaults()

    # Avoid failures in the tests for user-registered units
    import copy

    import brian2.units.fundamentalunits as fundamentalunits

    old_unit_registry = copy.copy(fundamentalunits.user_unit_register)
    fundamentalunits.user_unit_register = fundamentalunits.UnitRegistry()

    if float_dtype is not None:
        print(f"Setting dtype for floating point variables to: {float_dtype.__name__}")

        prefs["core.default_float_dtype"] = float_dtype

    print()

    # Suppress INFO log messages during testing
    from brian2.utils.logger import LOG_LEVELS, BrianLogger

    log_level = BrianLogger.console_handler.level
    BrianLogger.console_handler.setLevel(LOG_LEVELS["WARNING"])

    # Switch off code optimization to get faster compilation times
    prefs["codegen.cpp.extra_compile_args_gcc"].extend(["-w", "-O0"])
    prefs["codegen.cpp.extra_compile_args_msvc"].extend(["/Od"])

    pref_plugin = PreferencePlugin(prefs, fail_for_not_implemented)
    try:
        success = []
        pref_plugin.device = "runtime"
        pref_plugin.device_options = {}
        if test_codegen_independent:
            print("Running doctests")
            # Some doctests do actually use code generation, use numpy for that
            prefs["codegen.target"] = "numpy"
            # Always test doctests with 64 bit, to avoid differences in print output
            if float_dtype is not None:
                prefs["core.default_float_dtype"] = np.float64
            if sphinx_dir is None:
                sphinx_dir = os.path.abspath(
                    os.path.join(os.path.dirname(__file__), "..", "..", "docs_sphinx")
                )
            if os.path.exists(sphinx_dir):
                sphinx_doc_dir = [sphinx_dir]
            else:
                # When running on travis, the source directory is in the SRCDIR
                # environment variable
                if "SRCDIR" in os.environ:
                    sphinx_dir = os.path.abspath(
                        os.path.join(os.environ["SRCDIR"], "docs_sphinx")
                    )
                    if os.path.exists(sphinx_dir):
                        sphinx_doc_dir = [sphinx_dir]
                    else:
                        sphinx_doc_dir = []
                else:
                    sphinx_doc_dir = []
            argv = make_argv(dirnames + sphinx_doc_dir, doctests=True)
            if "codegen_independent" in test_in_parallel:
                argv.extend(multiprocess_arguments)
            success.append(
                pytest.main(argv + additional_args, plugins=[pref_plugin]) == 0
            )
            # Set float_dtype back again if necessary
            if float_dtype is not None:
                prefs["core.default_float_dtype"] = float_dtype

            print("Running tests that do not use code generation")
            argv = make_argv(dirnames, "codegen_independent", test_GSL=test_GSL)
            if "codegen_independent" in test_in_parallel:
                argv.extend(multiprocess_arguments)
            success.append(
                pytest.main(argv + additional_args, plugins=[pref_plugin]) == 0
            )
            clear_caches()

        for target in codegen_targets:
            print(f"Running tests for target {target}:")
            # Also set the target for string-expressions -- otherwise we'd only
            # ever test numpy for those
            prefs["codegen.target"] = target

            markers = "not standalone_only and not codegen_independent"
            if not long_tests:
                markers += " and not long"
            # explicitly ignore the brian2.hears file for testing, otherwise the
            # doctest search will import it, failing on Python 3
            argv = make_argv(dirnames, markers, test_GSL=test_GSL)
            if target in test_in_parallel:
                argv.extend(multiprocess_arguments)
            success.append(
                pytest.main(argv + additional_args, plugins=[pref_plugin]) == 0
            )
            clear_caches()

        pref_plugin.device = test_standalone
        if test_standalone:
            from brian2.devices.device import get_device, set_device

            pref_plugin.device_options = {"directory": None, "with_output": False}
            pref_plugin.device_options.update(build_options)
            print(f'Testing standalone device "{test_standalone}"')
            print("Running standalone-compatible standard tests (single run statement)")
            markers = "and not long" if not long_tests else ""
            markers += " and not multiple_runs"
            argv = make_argv(
                dirnames, f"standalone_compatible {markers}", test_GSL=test_GSL
            )
            if test_standalone in test_in_parallel:
                argv.extend(multiprocess_arguments)
            success.append(
                pytest.main(argv + additional_args, plugins=[pref_plugin]) in [0, 5]
            )
            clear_caches()

            reset_device()

            print(
                "Running standalone-compatible standard tests (multiple run statements)"
            )
            pref_plugin.device_options = {
                "directory": None,
                "with_output": False,
                "build_on_run": False,
            }
            pref_plugin.device_options.update(build_options)
            markers = " and not long" if not long_tests else ""
            markers += " and multiple_runs"
            argv = make_argv(
                dirnames, f"standalone_compatible{markers}", test_GSL=test_GSL
            )
            if test_standalone in test_in_parallel:
                argv.extend(multiprocess_arguments)
            success.append(
                pytest.main(argv + additional_args, plugins=[pref_plugin]) in [0, 5]
            )
            clear_caches()
            reset_device()

            if test_openmp and test_standalone == "cpp_standalone":
                # Run all the standalone compatible tests again with 4 threads
                pref_plugin.device_options = {"directory": None, "with_output": False}
                pref_plugin.device_options.update(build_options)
                prefs["devices.cpp_standalone.openmp_threads"] = 4
                print(
                    "Running standalone-compatible standard tests with OpenMP (single"
                    " run statements)"
                )
                markers = " and not long" if not long_tests else ""
                markers += " and not multiple_runs"
                argv = make_argv(
                    dirnames, f"standalone_compatible{markers}", test_GSL=test_GSL
                )
                success.append(
                    pytest.main(argv + additional_args, plugins=[pref_plugin]) in [0, 5]
                )
                clear_caches()
                reset_device()

                pref_plugin.device_options = {
                    "directory": None,
                    "with_output": False,
                    "build_on_run": False,
                }
                pref_plugin.device_options.update(build_options)
                print(
                    "Running standalone-compatible standard tests with OpenMP (multiple"
                    " run statements)"
                )
                markers = " and not long" if not long_tests else ""
                markers += " and multiple_runs"
                argv = make_argv(
                    dirnames, f"standalone_compatible{markers}", test_GSL=test_GSL
                )
                success.append(
                    pytest.main(argv + additional_args, plugins=[pref_plugin]) in [0, 5]
                )
                clear_caches()
                prefs["devices.cpp_standalone.openmp_threads"] = 0

                reset_device()

            print("Running standalone-specific tests")
            exclude_openmp = " and not openmp" if not test_openmp else ""
            argv = make_argv(
                dirnames, test_standalone + exclude_openmp, test_GSL=test_GSL
            )
            if test_standalone in test_in_parallel:
                argv.extend(multiprocess_arguments)
            success.append(
                pytest.main(argv + additional_args, plugins=[pref_plugin]) in [0, 5]
            )
            clear_caches()

        all_success = all(success)
        if not all_success:
            print(
                f"ERROR: {len(success) - sum(success)}/{len(success)} test suite(s) "
                "did not complete successfully (see above)."
            )
        else:
            print(
                f"OK: {len(success)}/{len(success)} test suite(s) did complete "
                "successfully."
            )
        return all_success

    finally:
        BrianLogger.console_handler.setLevel(log_level)

        if reset_preferences:
            # Restore the user preferences
            prefs.read_preference_file(StringIO(stored_prefs))
            prefs._backup()

        fundamentalunits.user_unit_register = old_unit_registry


if __name__ == "__main__":
    run()



"""
Module defining the `Network` object, the basis of all simulation runs.

Preferences
-----------

.. document_brian_prefs:: core.network

"""

import gc
import os
import pickle as pickle
import sys
import time
from collections import Counter, defaultdict, namedtuple
from collections.abc import Mapping, Sequence

from brian2.core.base import BrianObject, BrianObjectException
from brian2.core.clocks import Clock, defaultclock
from brian2.core.names import Nameable
from brian2.core.namespace import get_local_namespace
from brian2.core.preferences import BrianPreference, prefs
from brian2.devices.device import all_devices, get_device
from brian2.groups.group import Group
from brian2.synapses.synapses import SummedVariableUpdater
from brian2.units.allunits import msecond, second
from brian2.units.fundamentalunits import Quantity, check_units
from brian2.utils.logger import get_logger

from .base import device_override

__all__ = ["Network", "profiling_summary", "scheduling_summary"]


logger = get_logger(__name__)


prefs.register_preferences(
    "core.network",
    "Network preferences",
    default_schedule=BrianPreference(
        default=[
            "start",
            "groups",
            "thresholds",
            "synapses",
            "resets",
            "end",
        ],
        docs="""
        Default schedule used for networks that
        don't specify a schedule.
        """,
    ),
)


def _format_time(time_in_s):
    """
    Helper function to format time in seconds, minutes, hours, days, depending
    on the magnitude.

    Examples
    --------
    >>> from brian2.core.network import _format_time
    >>> _format_time(12345)
    '3h 25m 45s'
    >>> _format_time(123)
    '2m 3s'
    >>> _format_time(12.5)
    '12s'
    >>> _format_time(.5)
    '< 1s'

    """
    divisors = [24 * 60 * 60, 60 * 60, 60, 1]
    letters = ["d", "h", "m", "s"]
    remaining = time_in_s
    text = ""
    for divisor, letter in zip(divisors, letters):
        time_to_represent = int(remaining / divisor)
        remaining -= time_to_represent * divisor
        if time_to_represent > 0 or len(text):
            if len(text):
                text += " "
            text += f"{int(time_to_represent)}{letter}"

    # less than one second
    if len(text) == 0:
        text = "< 1s"

    return text


class TextReport:
    """
    Helper object to report simulation progress in `Network.run`.

    Parameters
    ----------
    stream : file
        The stream to write to, commonly `sys.stdout` or `sys.stderr`.
    """

    def __init__(self, stream):
        self.stream = stream

    def __call__(self, elapsed, completed, start, duration):
        if completed == 0.0:
            self.stream.write(
                f"Starting simulation at t={start} for a duration of {duration}\n"
            )
        else:
            t = str(completed * duration)
            percent = int(completed * 100.0)
            real_t = _format_time(float(elapsed))
            report_msg = f"{t} ({percent}%) simulated in {real_t}"
            if completed < 1.0:
                remaining = int(round((1 - completed) / completed * float(elapsed)))
                remaining_msg = f", estimated {_format_time(remaining)} remaining.\n"
            else:
                remaining_msg = "\n"

            self.stream.write(report_msg + remaining_msg)

        # Flush the stream, this is useful if stream is a file
        self.stream.flush()


def _format_table(header, values, cell_formats):
    # table = [header] + values
    table_format = len(values) * [cell_formats]
    col_widths = [
        max(len(format.format(cell, 0)) for format, cell in zip(col_format, col))
        for col_format, col in zip(
            list(zip(*([len(header) * ["{}"]] + table_format))),
            list(zip(*([header] + values))),
        )
    ]
    line = "-+-".join("-" * width for width in col_widths)
    content = [
        " | ".join(
            format.format(cell, width)
            for format, cell, width in zip(row_format, row, col_widths)
        )
        for row_format, row in zip(table_format, values)
    ]
    formatted_header = " | ".join(
        "{:^{}}".format(h, width) for h, width in zip(header, col_widths)
    )

    return "\n".join([formatted_header, line] + content)


class SchedulingSummary:
    """
    Object representing the schedule that is used to simulate the objects in a
    network. Objects of this type are returned by `scheduling_summary`, they
    should not be created manually by the user.

    Parameters
    ----------
    objects : list of `BrianObject`
        The sorted list of objects that are simulated by the network.
    """

    def __init__(self, objects):
        # Map each dt to a rank (i.e. smallest dt=0, second smallest=1, etc.)
        self.dts = {
            dt: rank
            for rank, dt in enumerate(sorted({float(obj.clock.dt) for obj in objects}))
        }
        ScheduleEntry = namedtuple(
            "ScheduleEntry",
            field_names=[
                "when",
                "order",
                "dt",
                "name",
                "type",
                "active",
                "owner_name",
                "owner_type",
            ],
        )
        entries = []
        for obj in objects:
            if len(obj.contained_objects):
                continue
            owner = getattr(obj, "group", None)
            if owner is None:
                owner_name, owner_type = None, None
            else:
                owner_name = owner.name
                owner_type = owner.__class__.__name__
            entries.append(
                ScheduleEntry(
                    when=obj.when,
                    order=obj.order,
                    dt=obj.clock.dt,
                    name=obj.name,
                    type=obj.__class__.__name__,
                    active=obj.active,
                    owner_name=owner_name,
                    owner_type=owner_type,
                )
            )
        self.entries = entries

        self.all_dts = sorted({float(entry.dt) for entry in self.entries})
        # How many steps compared to the fastest clock?
        self.steps = {float(dt): int(dt / self.all_dts[0]) for dt in self.all_dts}

    def __repr__(self):
        return _format_table(
            ["object", "part of", "Clock dt", "when", "order", "active"],
            [
                [
                    f"{entry.name} ({entry.type})",
                    (
                        f"{entry.owner_name} ({entry.owner_type})"
                        if entry.owner_name is not None
                        else "--"
                    ),
                    "{} (every {})".format(
                        str(entry.dt),
                        (
                            "step"
                            if self.steps[float(entry.dt)] == 1
                            else f"{self.steps[float(entry.dt)]} steps"
                        ),
                    ),
                    entry.when,
                    entry.order,
                    "yes" if entry.active else "no",
                ]
                for entry in self.entries
            ],
            ["{:<{}}", "{:<{}}", "{:<{}}", "{:<{}}", "{:{}d}", "{:^{}}"],
        )

    def _repr_html_(self):
        rows = [
            """\
        <tr>
            <td style="text-align: left;">{}</td>
            <td style="text-align: left;">{}</td>
            <td style="text-align: left;">{}</td>
            <td style="text-align: left;">{}</td>
            <td style="text-align: right;">{}</td>
            <td style="text-align: center;">{}</td>
        </tr>
        """.format(
                f"<b>{entry.name}</b> (<em>{entry.type}</em>)",
                (
                    f"{entry.owner_name} (<em>{entry.owner_type}</em>)"
                    if entry.owner_name is not None
                    else "&ndash;"
                ),
                "{} (every {})".format(
                    str(entry.dt),
                    (
                        "step"
                        if self.steps[float(entry.dt)] == 1
                        else f"{self.steps[float(entry.dt)]} steps"
                    ),
                ),
                entry.when,
                entry.order,
                "yes" if entry.active else "no",
            )
            for entry in self.entries
        ]
        html_code = """
        <table>
        <thead>
        <tr>
            <th style="text-align: center;">object</th>
            <th style="text-align: center;">part of</th>
            <th style="text-align: center;">Clock dt</th>
            <th style="text-align: center;">when</th>
            <th style="text-align: center;">order</th>
            <th style="text-align: center;">active</th>
        </tr>
        </thead>
        <tbody>
{rows}
        </tbody>
        </table>
        """.format(
            rows="\n".join(rows)
        )
        return html_code


def _check_multiple_summed_updaters(objects):
    """
    Helper function that checks whether multiple `SummedVariableUpdater` target
    the same target variable. Raises a `NotImplementedError` if this is the
    case (and problematic, i.e. not when using non-overlapping subgroups).

    Parameters
    ----------
    objects : list of `BrianObject`
        The list of objects in the network.
    """
    summed_targets = {}
    for obj in objects:
        if isinstance(obj, SummedVariableUpdater):
            if obj.target_var in summed_targets:
                other_target = summed_targets[obj.target_var]
                if obj.target == other_target:
                    # We raise an error, even though this could be ok in
                    # principle (e.g. two Synapses could target different
                    # subsets of the target groups, without using subgroups)
                    msg = (
                        "Multiple 'summed variables' target the "
                        f"variable '{obj.target_var.name}' in group "
                        f"'{obj.target.name}'. Use multiple variables in "
                        "the target group instead."
                    )
                    raise NotImplementedError(msg)
                elif (
                    obj.target.start < other_target.stop
                    and other_target.start < obj.target.stop
                ):
                    # Overlapping subgroups
                    msg = (
                        "Multiple 'summed variables' target the "
                        f"variable '{obj.target_var.name}' in overlapping "
                        f"groups '{other_target.name}' and '{obj.target.name}'. "
                        "Use separate variables in the target groups instead."
                    )
                    raise NotImplementedError(msg)
            summed_targets[obj.target_var] = obj.target


def _get_all_objects(objs):
    """
    Helper function to get all objects of a 'Network' along with their
    corresponding ``contained_objects``.

    Parameters
    ----------
    objs : Iterable
        List or set of objects

    Returns
    -------
    all_objects : set
        A set of all Network's objects and respective child objects.
    """
    all_objects = set()
    for obj in objs:
        all_objects.add(obj)
        all_objects |= _get_all_objects(obj.contained_objects)
    return all_objects


class Network(Nameable):
    """
    Network(*objs, name='network*')

    The main simulation controller in Brian

    `Network` handles the running of a simulation. It contains a set of Brian
    objects that are added with `~Network.add`. The `~Network.run` method
    actually runs the simulation. The main run loop, determining which
    objects get called in what order is described in detail in the notes below.
    The objects in the `Network` are accesible via their names, e.g.
    `net['neurongroup']` would return the `NeuronGroup` with this name.

    Parameters
    ----------
    objs : (`BrianObject`, container), optional
        A list of objects to be added to the `Network` immediately, see
        `~Network.add`.
    name : str, optional
        An explicit name, if not specified gives an automatically generated name

    Notes
    -----

    The main run loop performs the following steps:

    1. Prepare the objects if necessary, see `~Network.prepare`.
    2. Determine the end time of the simulation as `~Network.t`+``duration``.
    3. Determine which set of clocks to update. This will be the clock with the
       smallest value of `~Clock.t`. If there are several with the same value,
       then all objects with these clocks will be updated simultaneously.
       Set `~Network.t` to the clock time.
    4. If the `~Clock.t` value of these clocks is past the end time of the
       simulation, stop running. If the `Network.stop` method or the
       `stop` function have been called, stop running. Set `~Network.t` to the
       end time of the simulation.
    5. For each object whose `~BrianObject.clock` is set to one of the clocks from the
       previous steps, call the `~BrianObject.update` method. This method will
       not be called if the `~BrianObject.active` flag is set to ``False``.
       The order in which the objects are called is described below.
    6. Increase `Clock.t` by `Clock.dt` for each of the clocks and return to
       step 2.

    The order in which the objects are updated in step 4 is determined by
    the `Network.schedule` and the objects `~BrianObject.when` and
    `~BrianObject.order` attributes. The `~Network.schedule` is a list of
    string names. Each `~BrianObject.when` attribute should be one of these
    strings, and the objects will be updated in the order determined by the
    schedule. The default schedule is
    ``['start', 'groups', 'thresholds', 'synapses', 'resets', 'end']``. In
    addition to the names provided in the schedule, automatic names starting
    with ``before_`` and ``after_`` can be used. That means that all objects
    with ``when=='before_start'`` will be updated first, then
    those with ``when=='start'``, ``when=='after_start'``,
    ``when=='before_groups'``, ``when=='groups'`` and so forth. If several
    objects have the same `~BrianObject.when` attribute, then the order is
    determined by the `~BrianObject.order` attribute (lower first).

    See Also
    --------

    MagicNetwork, run, stop
    """

    def __init__(self, *objs, **kwds):
        #: The set of objects in the Network, should not normally be modified
        #: directly.
        #: Note that in a `MagicNetwork`, this attribute only contains the
        #: objects during a run: it is filled in `before_run` and emptied in
        #: `after_run`
        self.objects = set()

        name = kwds.pop("name", "network*")

        if kwds:
            raise TypeError("Only keyword argument to Network is 'name'.")

        Nameable.__init__(self, name=name)

        #: Current time as a float
        self.t_ = 0.0

        for obj in objs:
            self.add(obj)

        #: Stored state of objects (store/restore)
        self._stored_state = {}

        # Stored profiling information (if activated via the keyword option)
        self._profiling_info = None

        self._schedule = None

    t = property(
        fget=lambda self: Quantity(self.t_, dim=second.dim),
        doc="Current simulation time in seconds (`Quantity`)",
    )

    @device_override("network_get_profiling_info")
    def get_profiling_info(self):
        """
        The only reason this is not directly implemented in `profiling_info`
        is to allow devices (e.g. `CPPStandaloneDevice`) to overwrite this.
        """
        if self._profiling_info is None:
            raise ValueError(
                "No profiling info collected (did you run with 'profile=True?')"
            )
        return sorted(self._profiling_info, key=lambda item: item[1], reverse=True)

    @property
    def profiling_info(self):
        """
        The time spent in executing the various `CodeObject` s.

        A list of ``(name, time)`` tuples, containing the name of the
        `CodeObject` and the total execution time for simulations of this object
        (as a `Quantity` with unit `second`). The list is sorted descending
        with execution time.

        Profiling has to be activated using the ``profile`` keyword in `run` or
        `Network.run`.
        """
        return self.get_profiling_info()

    _globally_stopped = False
    _globally_running = False

    def __getitem__(self, item):
        if not isinstance(item, str):
            raise TypeError(
                f"Need a name to access objects in a Network, got {type(item)} instead"
            )
        all_objects = _get_all_objects(self.objects)
        for obj in all_objects:
            if obj.name == item:
                return obj

        raise KeyError(f'No object with name "{item}" found')

    def __delitem__(self, key):
        if not isinstance(key, str):
            raise TypeError(
                "Need a name to access objects in a Network, got {type(key)} instead"
            )
        for obj in self.objects:
            if obj.name == key:
                self.remove(obj)
                return

        raise KeyError(f"No object with name '{key}' found")

    def __contains__(self, item):
        all_objects = _get_all_objects(self.objects)
        for obj in all_objects:
            if obj.name == item:
                return True
        return False

    def __len__(self):
        all_objects = _get_all_objects(self.objects)
        return len(all_objects)

    def __iter__(self):
        all_objects = _get_all_objects(self.objects)
        return iter(all_objects)

    def add(self, *objs):
        """
        Add objects to the `Network`

        Parameters
        ----------

        objs : (`BrianObject`, container)
            The `BrianObject` or container of Brian objects to be added. Specify
            multiple objects, or lists (or other containers) of objects.
            Containers will be added recursively. If the container is a `dict`
            then it will add the values from the dictionary but not the keys.
            If you want to add the keys, do ``add(objs.keys())``.
        """
        for obj in objs:
            if isinstance(obj, BrianObject):
                if obj._network is not None:
                    raise RuntimeError(
                        f"{obj.name} has already been simulated, cannot "
                        "add it to the network. If you were "
                        "trying to remove and add an object to "
                        "temporarily stop it from being run, "
                        "set its active flag to False instead."
                    )
                self.objects.add(obj)
            else:
                # allow adding values from dictionaries
                if isinstance(obj, Mapping):
                    self.add(*list(obj.values()))
                else:
                    try:
                        for o in obj:
                            # The following "if" looks silly but avoids an infinite
                            # recursion if a string is provided as an argument
                            # (which might occur during testing)
                            if o is obj:
                                raise TypeError()
                            self.add(o)
                    except TypeError:
                        raise TypeError(
                            "Can only add objects of type BrianObject, "
                            "or containers of such objects to Network"
                        )

    def remove(self, *objs):
        """
        Remove an object or sequence of objects from a `Network`.

        Parameters
        ----------

        objs : (`BrianObject`, container)
            The `BrianObject` or container of Brian objects to be removed. Specify
            multiple objects, or lists (or other containers) of objects.
            Containers will be removed recursively.
        """
        for obj in objs:
            if isinstance(obj, BrianObject):
                self.objects.remove(obj)
            else:
                try:
                    for o in obj:
                        self.remove(o)
                except TypeError:
                    raise TypeError(
                        "Can only remove objects of type "
                        "BrianObject, or containers of such "
                        "objects from Network"
                    )

    def _full_state(self):
        all_objects = _get_all_objects(self.objects)
        state = {}
        for obj in all_objects:
            if hasattr(obj, "_full_state"):
                state[obj.name] = obj._full_state()
        clocks = {obj.clock for obj in all_objects}
        for clock in clocks:
            state[clock.name] = clock._full_state()
        # Store the time as "0_t" -- this name is guaranteed not to clash with
        # the name of an object as names are not allowed to start with a digit
        state["0_t"] = self.t_
        return state

    @device_override("network_store")
    def store(self, name="default", filename=None):
        """
        store(name='default', filename=None)

        Store the state of the network and all included objects.

        Parameters
        ----------
        name : str, optional
            A name for the snapshot, if not specified uses ``'default'``.
        filename : str, optional
            A filename where the state should be stored. If not specified, the
            state will be stored in memory.

        Notes
        -----
        The state stored to disk can be restored with the `Network.restore`
        function. Note that it will only restore the *internal state* of all
        the objects (including undelivered spikes) -- the objects have to
        exist already and they need to have the same name as when they were
        stored. Equations, thresholds, etc. are *not* stored -- this is
        therefore not a general mechanism for object serialization. Also, the
        format of the file is not guaranteed to work across platforms or
        versions. If you are interested in storing the state of a network for
        documentation or analysis purposes use `Network.get_states` instead.
        """
        clocks = {obj.clock for obj in _get_all_objects(self.objects)}
        # Make sure that all clocks are up to date
        for clock in clocks:
            clock._set_t_update_dt(target_t=self.t)

        state = self._full_state()
        # Store the state of the random number generator
        dev = get_device()
        state["_random_generator_state"] = dev.get_random_state()

        if filename is None:
            self._stored_state[name] = state
        else:
            # A single file can contain several states, so we'll read in the
            # existing file first if it exists
            if os.path.exists(filename):
                with open(filename, "rb") as f:
                    store_state = pickle.load(f)
            else:
                store_state = {}
            store_state[name] = state

            with open(filename, "wb") as f:
                pickle.dump(store_state, f, protocol=pickle.HIGHEST_PROTOCOL)

    @device_override("network_restore")
    def restore(self, name="default", filename=None, restore_random_state=False):
        """
        restore(name='default', filename=None, restore_random_state=False)

        Retore the state of the network and all included objects.

        Parameters
        ----------
        name : str, optional
            The name of the snapshot to restore, if not specified uses
            ``'default'``.
        filename : str, optional
            The name of the file from where the state should be restored. If
            not specified, it is expected that the state exist in memory
            (i.e. `Network.store` was previously called without the ``filename``
            argument).
        restore_random_state : bool, optional
            Whether to restore the state of the random number generator. If set
            to ``True``, going back to an earlier state of the simulation will
            continue exactly where it left off, even if the simulation is
            stochastic. If set to ``False`` (the default), random numbers are
            independent between runs (except for explicitly set random seeds),
            regardless of whether `store`/`restore` has been used or not. Note
            that this also restores numpy's random number generator (since it is
            used internally by Brian), but it does *not* restore Python's
            builtin random number generator in the ``random`` module.
        """
        all_objects = _get_all_objects(self.objects)
        if filename is None:
            state = self._stored_state[name]
        else:
            with open(filename, "rb") as f:
                state = pickle.load(f)[name]
        self.t_ = state["0_t"]
        if restore_random_state:
            dev = get_device()
            dev.set_random_state(state["_random_generator_state"])
        clocks = {obj.clock for obj in all_objects}
        restored_objects = set()
        for obj in all_objects:
            if obj.name in state:
                obj._restore_from_full_state(state[obj.name])
                restored_objects.add(obj.name)
            elif hasattr(obj, "_restore_from_full_state"):
                raise KeyError(
                    "Stored state does not have a stored state for "
                    f"'{obj.name}'. Note that the names of all objects have "
                    "to be identical to the names when they were "
                    "stored."
                )
        for clock in clocks:
            clock._restore_from_full_state(state[clock.name])
        clock_names = {c.name for c in clocks}

        unnused = (
            set(state.keys())
            - restored_objects
            - clock_names
            - {"0_t", "_random_generator_state"}
        )
        if len(unnused):
            raise KeyError(
                "The stored state contains the state of the "
                "following objects which were not present in the "
                f"network: {', '.join(unnused)}. Note that the names of all "
                "objects have to be identical to the names when they "
                "were stored."
            )

    def get_states(
        self,
        units=True,
        format="dict",
        subexpressions=False,
        read_only_variables=True,
        level=0,
    ):
        """
        Return a copy of the current state variable values of objects in the
        network.. The returned arrays are copies of the actual arrays that
        store the state variable values, therefore changing the values in the
        returned dictionary will not affect the state variables.

        Parameters
        ----------
        vars : list of str, optional
            The names of the variables to extract. If not specified, extract
            all state variables (except for internal variables, i.e. names that
            start with ``'_'``). If the ``subexpressions`` argument is ``True``,
            the current values of all subexpressions are returned as well.
        units : bool, optional
            Whether to include the physical units in the return value. Defaults
            to ``True``.
        format : str, optional
            The output format. Defaults to ``'dict'``.
        subexpressions: bool, optional
            Whether to return subexpressions when no list of variable names
            is given. Defaults to ``False``. This argument is ignored if an
            explicit list of variable names is given in ``vars``.
        read_only_variables : bool, optional
            Whether to return read-only variables (e.g. the number of neurons,
            the time, etc.). Setting it to ``False`` will assure that the
            returned state can later be used with `set_states`. Defaults to
            ``True``.
        level : int, optional
            How much higher to go up the stack to resolve external variables.
            Only relevant if extracting subexpressions that refer to external
            variables.

        Returns
        -------
        values : dict
            A dictionary mapping object names to the state variables of that
            object, in the specified ``format``.

        See Also
        --------
        VariableOwner.get_states
        """
        states = dict()
        for obj in self.sorted_objects:
            if hasattr(obj, "get_states"):
                states[obj.name] = obj.get_states(
                    vars=None,
                    units=units,
                    format=format,
                    subexpressions=subexpressions,
                    read_only_variables=read_only_variables,
                    level=level + 1,
                )
        return states

    def set_states(self, values, units=True, format="dict", level=0):
        """
        Set the state variables of objects in the network.

        Parameters
        ----------
        values : dict
            A dictionary mapping object names to objects of ``format``, setting
            the states of this object.
        units : bool, optional
            Whether the ``values`` include physical units. Defaults to ``True``.
        format : str, optional
            The format of ``values``. Defaults to ``'dict'``
        level : int, optional
            How much higher to go up the stack to _resolve external variables.
            Only relevant when using string expressions to set values.

        See Also
        --------
        Group.set_states
        """
        # For the moment, 'dict' is the only supported format -- later this will
        # be made into an extensible system, see github issue #306
        for obj_name, obj_values in values.items():
            if obj_name not in self:
                raise KeyError(
                    "Network does not include a network with name '%s'." % obj_name
                )
            self[obj_name].set_states(
                obj_values, units=units, format=format, level=level + 1
            )

    def _get_schedule(self):
        if self._schedule is None:
            return list(prefs.core.network.default_schedule)
        else:
            return list(self._schedule)

    def _set_schedule(self, schedule):
        if schedule is None:
            self._schedule = None
            logger.debug("Resetting network {self.name} schedule to default schedule")
        else:
            if not isinstance(schedule, Sequence) or not all(
                isinstance(slot, str) for slot in schedule
            ):
                raise TypeError(
                    "Schedule has to be None or a sequence of scheduling slots"
                )
            if any(
                slot.startswith("before_") or slot.startswith("after_")
                for slot in schedule
            ):
                raise ValueError(
                    "Slot names are not allowed to start with "
                    "'before_' or 'after_' -- such slot names "
                    "are created automatically based on the "
                    "existing slot names."
                )
            self._schedule = list(schedule)
            logger.debug(
                f"Setting network '{self.name}' schedule to {self._schedule}",
                "_set_schedule",
            )

    schedule = property(
        fget=_get_schedule,
        fset=_set_schedule,
        doc="""
        List of ``when`` slots in the order they will be updated, can be modified.

        See notes on scheduling in `Network`. Note that additional ``when``
        slots can be added, but the schedule should contain at least all of the
        names in the default schedule:
        ``['start', 'groups', 'thresholds', 'synapses', 'resets', 'end']``.

        The schedule can also be set to ``None``, resetting it to the default
        schedule set by the `core.network.default_schedule` preference.
        """,
    )

    @property
    def sorted_objects(self):
        """
        The sorted objects of this network in the order defined by the schedule.

        Objects are sorted first by their ``when`` attribute, and secondly
        by the ``order`` attribute. The order of the ``when`` attribute is
        defined by the ``schedule``. In addition to the slot names defined in
        the schedule, automatic slot names starting with ``before_`` and
        ``after_`` can be used (e.g. the slots ``['groups', 'thresholds']``
        allow to use ``['before_groups', 'groups', 'after_groups',
        'before_thresholds', 'thresholds', 'after_thresholds']``).

        Final ties are resolved using the objects' names, leading to an
        arbitrary but deterministic sorting.
        """
        # Provided slot names are assigned positions 1, 4, 7, ...
        # before_... names are assigned positions 0, 3, 6, ...
        # after_... names are assigned positions 2, 5, 8, ...
        all_objects = _get_all_objects(self.objects)
        when_to_int = {when: 1 + i * 3 for i, when in enumerate(self.schedule)}
        when_to_int.update(
            (f"before_{when}", i * 3) for i, when in enumerate(self.schedule)
        )
        when_to_int.update(
            (f"after_{when}", 2 + i * 3) for i, when in enumerate(self.schedule)
        )
        return sorted(
            all_objects, key=lambda obj: (when_to_int[obj.when], obj.order, obj.name)
        )

    def scheduling_summary(self):
        """
        Return a `SchedulingSummary` object, representing the scheduling
        information for all objects included in the network.

        Returns
        -------
        summary : `SchedulingSummary`
            Object representing the scheduling information.
        """
        return SchedulingSummary(self.sorted_objects)

    def check_dependencies(self):
        all_objects = _get_all_objects(self.objects)
        all_ids = [obj.id for obj in all_objects]
        for obj in all_objects:
            for dependency in obj._dependencies:
                if dependency not in all_ids:
                    raise ValueError(
                        f"'{obj.name}' has been included in the network "
                        "but not the object on which it "
                        "depends."
                    )

    @device_override("network_before_run")
    def before_run(self, run_namespace):
        """
        before_run(namespace)

        Prepares the `Network` for a run.

        Objects in the `Network` are sorted into the correct running order, and
        their `BrianObject.before_run` methods are called.

        Parameters
        ----------
        run_namespace : dict-like, optional
            A namespace in which objects which do not define their own
            namespace will be run.
        """
        all_objects = self.sorted_objects
        prefs.check_all_validated()

        # Check names in the network for uniqueness
        names = [obj.name for obj in all_objects]
        non_unique_names = [name for name, count in Counter(names).items() if count > 1]
        if len(non_unique_names):
            formatted_names = ", ".join(f"'{name}'" for name in non_unique_names)
            raise ValueError(
                "All objects in a network need to have unique "
                "names, the following name(s) were used more than "
                f"once: {formatted_names}"
            )

        # Check that there are no SummedVariableUpdaters targeting the same
        # target variable
        _check_multiple_summed_updaters(all_objects)

        self._stopped = False
        Network._globally_stopped = False

        device = get_device()
        if device.network_schedule is not None:
            # The device defines a fixed network schedule
            if device.network_schedule != self.schedule:
                # TODO: The human-readable name of a device should be easier to get
                device_name = list(all_devices.keys())[
                    list(all_devices.values()).index(device)
                ]
                logger.warn(
                    f"The selected device '{device_name}' only "
                    "supports a fixed schedule, but this schedule is "
                    "not consistent with the network's schedule. The "
                    "simulation will use the device's schedule.\n"
                    f"Device schedule: {device.network_schedule}\n"
                    f"Network schedule: {self.schedule}\n"
                    "Set the network schedule explicitly or set the "
                    "core.network.default_schedule preference to "
                    "avoid this warning.",
                    name_suffix="schedule_conflict",
                    once=True,
                )

        objnames = ", ".join(obj.name for obj in all_objects)
        logger.debug(
            f"Preparing network '{self.name}' with {len(all_objects)} "
            f"objects: {objnames}",
            "before_run",
        )

        self.check_dependencies()

        for obj in all_objects:
            if obj.active:
                try:
                    obj.before_run(run_namespace)
                except Exception as ex:
                    raise BrianObjectException(
                        "An error occurred when preparing an object.", obj
                    ) from ex

        # Check that no object has been run as part of another network before
        for obj in all_objects:
            if obj._network is None:
                obj._network = self.id
            elif obj._network != self.id:
                raise RuntimeError(
                    f"'{obj.name}' has already been run in the "
                    "context of another network. Use "
                    "add/remove to change the objects "
                    "in a simulated network instead of "
                    "creating a new one."
                )

        clocknames = ", ".join(f"{obj.name} (dt={obj.dt})" for obj in self._clocks)
        logger.debug(
            f"Network '{self.name}' uses {len(self._clocks)} clocks: {clocknames}",
            "before_run",
        )

    @device_override("network_after_run")
    def after_run(self):
        """
        after_run()
        """
        for obj in self.sorted_objects:
            if obj.active:
                obj.after_run()

    def _nextclocks(self):
        clocks_times_dt = [
            (c, self._clock_variables[c][1][0], self._clock_variables[c][2][0])
            for c in self._clocks
        ]
        minclock, min_time, minclock_dt = min(clocks_times_dt, key=lambda k: k[1])
        curclocks = {
            clock
            for clock, time, dt in clocks_times_dt
            if (
                time == min_time
                or abs(time - min_time) / min(minclock_dt, dt) < Clock.epsilon_dt
            )
        }
        return minclock, curclocks

    @device_override("network_run")
    @check_units(duration=second, report_period=second)
    def run(
        self,
        duration,
        report=None,
        report_period=10 * second,
        namespace=None,
        profile=None,
        level=0,
    ):
        """
        run(duration, report=None, report_period=60*second, namespace=None, level=0)

        Runs the simulation for the given duration.

        Parameters
        ----------
        duration : `Quantity`
            The amount of simulation time to run for.
        report : {None, 'text', 'stdout', 'stderr', function}, optional
            How to report the progress of the simulation. If ``None``, do not
            report progress. If ``'text'`` or ``'stdout'`` is specified, print
            the progress to stdout. If ``'stderr'`` is specified, print the
            progress to stderr. Alternatively, you can specify a callback
            ``callable(elapsed, completed, start, duration)`` which will be passed
            the amount of time elapsed as a `Quantity`, the fraction ``completed``
            from 0.0 to 1.0, the ``start`` time of the simulation as a `Quantity`
            and the total duration of the simulation (in biological time)
            as a `Quantity`.
            The function will always be called at the beginning and the end
            (i.e. for fractions 0.0 and 1.0), regardless of the ``report_period``.
        report_period : `Quantity`
            How frequently (in real time) to report progress.
        namespace : dict-like, optional
            A namespace that will be used in addition to the group-specific
            namespaces (if defined). If not specified, the locals
            and globals around the run function will be used.
        profile : bool, optional
            Whether to record profiling information (see
            `Network.profiling_info`). Defaults to ``None`` (which will use the
            value set by ``set_device``, if any).
        level : int, optional
            How deep to go up the stack frame to look for the locals/global
            (see `namespace` argument). Only used by run functions that call
            this run function, e.g. `MagicNetwork.run` to adjust for the
            additional nesting.

        Notes
        -----
        The simulation can be stopped by calling `Network.stop` or the
        global `stop` function.

        Raises
        ------
        ValueError
            Error raised when duration passed in run function is negative.
        """
        if duration < 0:
            raise ValueError(
                f"Function 'run' expected a non-negative duration but got '{duration}'"
            )
        # This will trigger warnings for objects that have not been included in a network
        gc.collect()
        device = get_device()  # Do not use the ProxyDevice -- slightly faster

        if profile is None:
            profile = device.build_options.get("profile", False)

        all_objects = self.sorted_objects
        self._clocks = {obj.clock for obj in all_objects}
        single_clock = len(self._clocks) == 1

        t_start = self.t
        t_end = self.t + duration

        if single_clock:
            clock = list(self._clocks)[0]
            clock.set_interval(self.t, t_end)
        else:
            # We get direct references to the underlying variables for all clocks
            # to avoid expensive access during the run loop
            self._clock_variables = {
                c: (
                    c.variables["timestep"].get_value(),
                    c.variables["t"].get_value(),
                    c.variables["dt"].get_value(),
                )
                for c in self._clocks
            }
            for clock in self._clocks:
                clock.set_interval(self.t, t_end)

        # Get the local namespace
        if namespace is None:
            namespace = get_local_namespace(level=level + 3)

        self.before_run(namespace)

        if len(all_objects) == 0:
            return  # TODO: raise an error? warning?

        start_time = time.time()

        logger.debug(
            f"Simulating network '{self.name}' from time {t_start} to {t_end}.", "run"
        )

        if report is not None:
            report_period = float(report_period)
            next_report_time = start_time + report_period
            if report == "text" or report == "stdout":
                report_callback = TextReport(sys.stdout)
            elif report == "stderr":
                report_callback = TextReport(sys.stderr)
            elif isinstance(report, str):
                raise ValueError(
                    f'Do not know how to handle report argument "{report}".'
                )
            elif callable(report):
                report_callback = report
            else:
                raise TypeError(
                    "Do not know how to handle report argument, "
                    "it has to be one of 'text', 'stdout', "
                    "'stderr', or a callable function/object, "
                    f"but it is of type {type(report)}"
                )
            report_callback(0 * second, 0.0, t_start, duration)

        profiling_info = defaultdict(float)

        if single_clock:
            timestep, t, dt = (
                clock.variables["timestep"].get_value(),
                clock.variables["t"].get_value(),
                clock.variables["dt"].get_value(),
            )
        else:
            # Find the first clock to be updated (see note below)
            clock, curclocks = self._nextclocks()
            timestep, _, _ = self._clock_variables[clock]

        running = timestep[0] < clock._i_end

        active_objects = [obj for obj in all_objects if obj.active]

        Network._globally_running = True
        while running and not self._stopped and not Network._globally_stopped:
            if not single_clock:
                timestep, t, dt = self._clock_variables[clock]
            # update the network time to this clock's time
            self.t_ = t[0]
            if report is not None:
                current = time.time()
                if current > next_report_time:
                    report_callback(
                        (current - start_time) * second,
                        (self.t_ - float(t_start)) / float(t_end - t_start),
                        t_start,
                        duration,
                    )
                    next_report_time = current + report_period

            # update the objects and tick forward the clock(s)
            if single_clock:
                if profile:
                    for obj in active_objects:
                        obj_time = time.time()
                        obj.run()
                        profiling_info[obj.name] += time.time() - obj_time
                else:
                    for obj in active_objects:
                        obj.run()

                timestep[0] += 1
                t[0] = timestep[0] * dt[0]
            else:
                if profile:
                    for obj in active_objects:
                        if obj._clock in curclocks:
                            obj_time = time.time()
                            obj.run()
                            profiling_info[obj.name] += time.time() - obj_time
                else:
                    for obj in active_objects:
                        if obj._clock in curclocks:
                            obj.run()

                for c in curclocks:
                    timestep, t, dt = self._clock_variables[c]
                    timestep[0] += 1
                    t[0] = timestep[0] * dt[0]
                # find the next clocks to be updated. The < operator for Clock
                # determines that the first clock to be updated should be the one
                # with the smallest t value, unless there are several with the
                # same t value in which case we update all of them
                clock, curclocks = self._nextclocks()
                timestep, _, _ = self._clock_variables[clock]

            if (
                device._maximum_run_time is not None
                and time.time() - start_time > float(device._maximum_run_time)
            ):
                self._stopped = True
            else:
                running = timestep[0] < clock._i_end

        end_time = time.time()
        Network._globally_running = False
        if self._stopped or Network._globally_stopped:
            self.t_ = clock.t_
        else:
            self.t_ = float(t_end)

        device._last_run_time = end_time - start_time
        if duration > 0:
            device._last_run_completed_fraction = (self.t - t_start) / duration
        else:
            device._last_run_completed_fraction = 1.0

        # check for nans
        for obj in all_objects:
            if isinstance(obj, Group):
                obj._check_for_invalid_states()

        if report is not None:
            report_callback(
                (end_time - start_time) * second,
                device._last_run_completed_fraction,
                t_start,
                duration,
            )
        self.after_run()

        logger.debug(
            f"Finished simulating network '{self.name}' "
            f"(took {end_time-start_time:.2f}s)",
            "run",
        )
        # Store profiling info (or erase old info to avoid confusion)
        if profile:
            self._profiling_info = [
                (name, t * second) for name, t in profiling_info.items()
            ]
            # Dump a profiling summary to the log
            logger.debug(f"\n{str(profiling_summary(self))}")
        else:
            self._profiling_info = None

    @device_override("network_stop")
    def stop(self):
        """
        stop()

        Stops the network from running, this is reset the next time `Network.run` is called.
        """
        self._stopped = True

    def __repr__(self):
        objects = ", ".join(obj.__repr__() for obj in _get_all_objects(self.objects))
        return (
            f"<{self.__class__.__name__} at time t={self.t!s}, containing "
            f"objects: {objects}>"
        )


class ProfilingSummary:
    """
    Class to nicely display the results of profiling. Objects of this class are
    returned by `profiling_summary`.

    Parameters
    ----------

    net : `Network`
        The `Network` object to profile.
    show : int, optional
        The number of results to show (the longest results will be shown). If
        not specified, all results will be shown.

    See Also
    --------
    Network.profiling_info
    """

    def __init__(self, net, show=None):
        prof = net.profiling_info
        if len(prof):
            names, times = list(zip(*prof))
        else:  # Can happen if a network has been run for 0ms
            # Use a dummy entry to prevent problems with empty lists later
            names = ["no code objects have been run"]
            times = [0 * second]
        self.total_time = sum(times)
        self.time_unit = msecond
        if self.total_time > 1 * second:
            self.time_unit = second
        if show is not None:
            names = names[:show]
            times = times[:show]
        if self.total_time > 0 * second:
            self.percentages = [100.0 * time / self.total_time for time in times]
        else:
            self.percentages = [0.0 for _ in times]
        self.names_maxlen = max(len(name) for name in names)
        self.names = [name + " " * (self.names_maxlen - len(name)) for name in names]
        self.times = times

    def __repr__(self):
        times = [f"{time / self.time_unit:.2f} {self.time_unit}" for time in self.times]
        times_maxlen = max(len(time) for time in times)
        times = [" " * (times_maxlen - len(time)) + time for time in times]
        percentages = [f"{percentage:.2f} %" for percentage in self.percentages]
        percentages_maxlen = max(len(percentage) for percentage in percentages)
        percentages = [
            (" " * (percentages_maxlen - len(percentage))) + percentage
            for percentage in percentages
        ]

        s = "Profiling summary"
        s += f"\n{'=' * len(s)}\n"
        for name, t, percentage in zip(self.names, times, percentages):
            s += f"{name}    {t}    {percentage}\n"
        return s

    def _repr_html_(self):
        times = [f"{time / self.time_unit:.2f} {self.time_unit}" for time in self.times]
        percentages = [f"{percentage:.2f} %" for percentage in self.percentages]
        s = '<h2 class="brian_prof_summary_header">Profiling summary</h2>\n'
        s += '<table class="brian_prof_summary_table">\n'
        for name, t, percentage in zip(self.names, times, percentages):
            s += "<tr>"
            s += f"<td>{name}</td>"
            s += f'<td style="text-align: right">{t}</td>'
            s += f'<td style="text-align: right">{percentage}</td>'
            s += "</tr>\n"
        s += "</table>"
        return s


def profiling_summary(net=None, show=None):
    """
    Returns a `ProfilingSummary` of the profiling info for a run. This object
    can be transformed to a string explicitly but on an interactive console
    simply calling `profiling_summary` is enough since it will
    automatically convert the `ProfilingSummary` object.

    Parameters
    ----------

    net : {`Network`, None} optional
        The `Network` object to profile, or `magic_network` if not specified.
    show : int
        The number of results to show (the longest results will be shown). If
        not specified, all results will be shown.
    """
    if net is None:
        from .magic import magic_network

        net = magic_network
    return ProfilingSummary(net, show)


def scheduling_summary(net=None):
    """
    Returns a `SchedulingSummary` object, representing the scheduling
    information for all objects included in the given `Network` (or the
    "magic" network, if none is specified). The returned objects can be
    printed or converted to a string to give an ASCII table representation of
    the schedule. In a Jupyter notebook, the output can be displayed as a
    HTML table.

    Parameters
    ----------
    net : `Network`, optional
        The network for which the scheduling information should be displayed.
        Defaults to the "magic" network.

    Returns
    -------
    summary : `SchedulingSummary`
        An object that represents the scheduling information.
    """
    if net is None:
        from .magic import magic_network

        magic_network._update_magic_objects(level=1)
        net = magic_network
    return net.scheduling_summary()


def schedule_propagation_offset(net=None):
    """
    Returns the minimal time difference for a post-synaptic effect after a
    spike. With the default schedule, this time difference is 0, since the
    ``thresholds`` slot precedes the ``synapses`` slot. For the GeNN device,
    however, a post-synaptic effect will occur in the following time step, this
    function therefore returns one ``dt``.

    Parameters
    ----------
    net : `Network`
        The network to check (uses the magic network if not specified).

    Returns
    -------
    offset : `Quantity`
        The minimum spike propagation delay: ``0*ms`` for the standard schedule
        but ``dt`` for schedules where ``synapses`` precedes ``thresholds``.

    Notes
    -----
    This function always returns ``0*ms`` or ``defaultclock.dt`` -- no attempt
    is made to deal with other clocks.
    """
    from brian2.core.magic import magic_network

    device = get_device()
    if device.network_schedule is not None:
        schedule = device.network_schedule
    else:
        if net is None:
            net = magic_network
        schedule = net.schedule

    if schedule.index("thresholds") < schedule.index("synapses"):
        return 0 * second
    else:
        return defaultclock.dt



import re
import uuid

from brian2.core.tracking import Trackable
from brian2.utils.logger import get_logger

__all__ = ["Nameable"]

logger = get_logger(__name__)


def find_name(name, names=None):
    """
    Determine a unique name. If the desired ``name`` is already taken, will try
    to use a derived ``name_1``, ``name_2``, etc.

    Parameters
    ----------
    name : str
        The desired name.
    names : Iterable, optional
        A set of names that are already taken. If not provided, will use the names
        of all Brian objects as stored in `Nameable`.

    Returns
    -------
    unique_name : str
        A name based on ``name`` or ``name`` itself, unique with respect to the
        names in ``names``.
    """
    if not name.endswith("*"):
        # explicitly given names are used as given. Network.before_run (and
        # the device in case of standalone) will check for name clashes later
        return name

    name = name[:-1]

    if names is None:
        instances = set(Nameable.__instances__())
        allnames = {obj().name for obj in instances if hasattr(obj(), "name")}
    else:
        allnames = names

    # Try the name without any additions first:
    if name not in allnames:
        return name

    # Name is already taken, try _1, _2, etc.
    i = 1
    while f"{name}_{str(i)}" in allnames:
        i += 1
    return f"{name}_{str(i)}"


class Nameable(Trackable):
    """
    Base class to find a unique name for an object

    If you specify a name explicitly, and it has already been taken, a
    `ValueError` is raised. You can also specify a name with a wildcard asterisk
    in the end, i.e. in the form ``'name*'``. It will then try ``name`` first
    but if this is already specified, it will try ``name_1``, `name__2``, etc.
    This is the default mechanism used by most core objects in Brian, e.g.
    `NeuronGroup` uses a default name of ``'neurongroup*'``.

    Parameters
    ----------
    name : str
        An name for the object, possibly ending in ``*`` to specify that
        variants of this name should be tried if the name (without the asterisk)
        is already taken. If (and only if) the name for this object has already
        been set, it is also possible to call the initialiser with ``None`` for
        the `name` argument. This situation can arise when a class derives from
        multiple classes that derive themselves from `Nameable` (e.g. `Group`
        and `CodeRunner`) and their initialisers are called explicitely.

    Raises
    ------
    ValueError
        If the name is already taken.
    """

    def __init__(self, name):
        if getattr(self, "_name", None) is not None and name is None:
            # name has already been specified previously
            return

        self.assign_id()

        if not isinstance(name, str):
            raise TypeError(
                "'name' argument has to be a string, is type "
                f"{repr(type(name))} instead"
            )
        if not re.match(r"[_A-Za-z][_a-zA-Z0-9]*\*?$", name):
            raise ValueError(f"Name {name} not valid variable name")

        self._name = find_name(name)
        logger.diagnostic(
            f"Created object of class {self.__class__.__name__} with name {self._name}"
        )

    def assign_id(self):
        """
        Assign a new id to this object. Under most circumstances, this method
        should only be called once at the creation of the object to generate a
        unique id. In the case of the `MagicNetwork`, however, the id should
        change when a new, independent set of objects is simulated.
        """
        self._id = uuid.uuid4()

    name = property(
        fget=lambda self: self._name,
        doc="""
        The unique name for this object.

        Used when generating code. Should be an acceptable
        variable name, i.e. starting with a letter
        character and followed by alphanumeric characters and
        ``_``.
        """,
    )

    id = property(
        fget=lambda self: self._id,
        doc="""
        A unique id for this object.

        In contrast to names, which may be reused, the id stays
        unique. This is used in the dependency checking to not
        have to deal with the chore of comparing weak
        references, weak proxies and strong references.
        """,
    )



import os
import shutil
import tempfile

import pytest
from numpy.testing import assert_equal

from brian2 import *
from brian2.codegen.codeobject import CodeObject
from brian2.codegen.cpp_prefs import compiler_supports_c99
from brian2.codegen.generators import CodeGenerator
from brian2.core.functions import timestep
from brian2.devices import RuntimeDevice
from brian2.parsing.sympytools import str_to_sympy, sympy_to_str
from brian2.tests.utils import assert_allclose, exc_isinstance
from brian2.utils.logger import catch_logs


@pytest.mark.codegen_independent
def test_constants_sympy():
    """
    Make sure that symbolic constants are understood correctly by sympy
    """
    assert sympy_to_str(str_to_sympy("1.0/inf")) == "0"
    assert sympy_to_str(str_to_sympy("sin(pi)")) == "0"
    assert sympy_to_str(str_to_sympy("log(e)")) == "1"


@pytest.mark.standalone_compatible
def test_constants_values():
    """
    Make sure that symbolic constants use the correct values in code
    """
    G = NeuronGroup(3, "v : 1")
    G.v[0] = "pi"
    G.v[1] = "e"
    G.v[2] = "inf"
    run(0 * ms)
    assert_allclose(G.v[:], [np.pi, np.e, np.inf])


def int_(x):
    return array(x, dtype=int)


int_.__name__ = "int"


@pytest.mark.parametrize(
    "func,needs_c99_support",
    [
        (cos, False),
        (tan, False),
        (sinh, False),
        (cosh, False),
        (tanh, False),
        (arcsin, False),
        (arccos, False),
        (arctan, False),
        (log, False),
        (log10, False),
        (log1p, True),
        (exp, False),
        (np.sqrt, False),
        (expm1, True),
        (exprel, True),
        (np.ceil, False),
        (np.floor, False),
        (np.sign, False),
        (int_, False),
    ],
)
@pytest.mark.standalone_compatible
def test_math_functions(func, needs_c99_support):
    """
    Test that math functions give the same result, regardless of whether used
    directly or in generated Python or C++ code.
    """
    if not get_device() == RuntimeDevice or prefs.codegen.target != "numpy":
        if needs_c99_support and not compiler_supports_c99():
            pytest.skip('Support for function "{}" needs a compiler with C99 support.')
    test_array = np.array([-1, -0.5, 0, 0.5, 1])

    with catch_logs() as _:  # Let's suppress warnings about illegal values
        # Calculate the result directly
        numpy_result = func(test_array)

        # Calculate the result in a somewhat complicated way by using a
        # subexpression in a NeuronGroup
        if func.__name__ == "absolute":
            # we want to use the name abs instead of absolute
            func_name = "abs"
        else:
            func_name = func.__name__
        G = NeuronGroup(
            len(test_array),
            f"""func = {func_name}(variable) : 1
                           variable : 1""",
        )
        G.variable = test_array
        mon = StateMonitor(G, "func", record=True)
        net = Network(G, mon)
        net.run(defaultclock.dt)

        assert_allclose(
            numpy_result,
            mon.func_.flatten(),
            err_msg=f"Function {func.__name__} did not return the correct values",
        )


@pytest.mark.standalone_compatible
@pytest.mark.parametrize("func,operator", [(np.power, "**"), (np.mod, "%")])
def test_math_operators(func, operator):
    default_dt = defaultclock.dt
    test_array = np.array([-1, -0.5, 0, 0.5, 1])
    # Functions/operators
    scalar = 3

    # Calculate the result directly
    numpy_result = func(test_array, scalar)

    # Calculate the result in a somewhat complicated way by using a
    # subexpression in a NeuronGroup
    G = NeuronGroup(
        len(test_array),
        f"""func = variable {operator} scalar : 1
                       variable : 1""",
    )
    G.variable = test_array
    mon = StateMonitor(G, "func", record=True)
    net = Network(G, mon)
    net.run(defaultclock.dt)

    assert_allclose(
        numpy_result,
        mon.func_.flatten(),
        err_msg=f"Function {func.__name__} did not return the correct values",
    )


@pytest.mark.standalone_compatible
def test_clip():
    G = NeuronGroup(
        4,
        """
        clipexpr1 = clip(integer_var1, 0, 1) : integer
        clipexpr2 = clip(integer_var2, -0.5, 1.5) : integer
        clipexpr3 = clip(float_var1, 0, 1) : 1
        clipexpr4 = clip(float_var2, -0.5, 1.5) : 1
        integer_var1 : integer
        integer_var2 : integer
        float_var1 : 1
        float_var2 : 1
        """,
    )
    G.integer_var1 = [0, 1, -1, 2]
    G.integer_var2 = [0, 1, -1, 2]
    G.float_var1 = [0.0, 1.0, -1.0, 2.0]
    G.float_var2 = [0.0, 1.0, -1.0, 2.0]
    s_mon = StateMonitor(
        G, ["clipexpr1", "clipexpr2", "clipexpr3", "clipexpr4"], record=True
    )
    run(defaultclock.dt)
    assert_equal(s_mon.clipexpr1.flatten(), [0, 1, 0, 1])
    assert_equal(s_mon.clipexpr2.flatten(), [0, 1, 0, 1])
    assert_allclose(s_mon.clipexpr3.flatten(), [0, 1, 0, 1])
    assert_allclose(s_mon.clipexpr4.flatten(), [0, 1, -0.5, 1.5])


@pytest.mark.standalone_compatible
def test_bool_to_int():
    # Test that boolean expressions and variables are correctly converted into
    # integers
    G = NeuronGroup(
        2,
        """
        intexpr1 = int(bool_var) : integer
        intexpr2 = int(float_var > 1.0) : integer
        bool_var : boolean
        float_var : 1
        """,
    )
    G.bool_var = [True, False]
    G.float_var = [2.0, 0.5]
    s_mon = StateMonitor(G, ["intexpr1", "intexpr2"], record=True)
    run(defaultclock.dt)
    assert_equal(s_mon.intexpr1.flatten(), [1, 0])
    assert_equal(s_mon.intexpr2.flatten(), [1, 0])


@pytest.mark.standalone_compatible
def test_integer_power():
    # See github issue #1500
    G = NeuronGroup(
        3,
        """
        intval1 : integer
        intval2 : integer
        k : integer (constant)
        """,
    )
    G.k = [0, 1, 2]
    G.run_regularly("intval1 = 2**k; intval2 = (-1)**k")
    run(defaultclock.dt)
    assert_equal(G.intval1[:], [1, 2, 4])
    assert_equal(G.intval2[:], [1, -1, 1])


@pytest.mark.codegen_independent
def test_timestep_function():
    dt = defaultclock.dt_
    # Check that multiples of dt end up in the correct time step
    t = np.arange(100000) * dt
    assert_equal(timestep(t, dt), np.arange(100000))

    # Scalar values should stay scalar
    ts = timestep(0.0005, 0.0001)
    assert np.isscalar(ts) and ts == 5

    # Length-1 arrays should stay arrays
    ts = timestep(np.array([0.0005]), 0.0001)
    assert ts.shape == (1,) and ts == 5


@pytest.mark.standalone_compatible
def test_timestep_function_during_run():
    group = NeuronGroup(
        2,
        """ref_t : second
                              ts = timestep(ref_t, dt) + timestep(t, dt) : integer""",
    )
    group.ref_t = [-1e4 * second, 5 * defaultclock.dt]
    mon = StateMonitor(group, "ts", record=True)
    run(5 * defaultclock.dt)
    assert all(mon.ts[0] <= -1e4)
    assert_equal(mon.ts[1], [5, 6, 7, 8, 9])


@pytest.mark.standalone_compatible
def test_user_defined_function():
    @implementation(
        "cpp",
        """
        inline double usersin(double x)
        {
            return sin(x);
        }
        """,
    )
    @implementation(
        "cython",
        """
        cdef double usersin(double x):
            return sin(x)
        """,
    )
    @check_units(x=1, result=1)
    def usersin(x):
        return np.sin(x)

    default_dt = defaultclock.dt
    test_array = np.array([0, 1, 2, 3])
    G = NeuronGroup(
        len(test_array),
        """
        func = usersin(variable) : 1
        variable : 1
        """,
    )
    G.variable = test_array
    mon = StateMonitor(G, "func", record=True)
    run(default_dt)
    assert_allclose(np.sin(test_array), mon.func_.flatten())


def test_user_defined_function_units():
    """
    Test the preparation of functions for use in code with check_units.
    """
    prefs.codegen.target = "numpy"
    if prefs.codegen.target != "numpy":
        pytest.skip("numpy-only test")

    def nothing_specified(x, y, z):
        return x * (y + z)

    no_result_unit = check_units(x=1, y=second, z=second)(nothing_specified)
    all_specified = check_units(x=1, y=second, z=second, result=second)(
        nothing_specified
    )
    consistent_units = check_units(x=None, y=None, z="y", result=lambda x, y, z: x * y)(
        nothing_specified
    )

    G = NeuronGroup(
        1,
        """
        a : 1
        b : second
        c : second
        """,
        namespace={
            "nothing_specified": nothing_specified,
            "no_result_unit": no_result_unit,
            "all_specified": all_specified,
            "consistent_units": consistent_units,
        },
    )
    net = Network(G)
    net.run(0 * ms)  # make sure we have a clock and therefore a t
    G.c = "all_specified(a, b, t)"
    G.c = "consistent_units(a, b, t)"
    with pytest.raises(ValueError):
        setattr(G, "c", "no_result_unit(a, b, t)")
    with pytest.raises(KeyError):
        setattr(G, "c", "nothing_specified(a, b, t)")
    with pytest.raises(DimensionMismatchError):
        setattr(G, "a", "all_specified(a, b, t)")
    with pytest.raises(DimensionMismatchError):
        setattr(G, "a", "all_specified(b, a, t)")
    with pytest.raises(DimensionMismatchError):
        setattr(G, "a", "consistent_units(a, b, t)")
    with pytest.raises(DimensionMismatchError):
        setattr(G, "a", "consistent_units(b, a, t)")


def test_simple_user_defined_function():
    # Make sure that it's possible to use a Python function directly, without
    # additional wrapping
    @check_units(x=1, result=1)
    def usersin(x):
        return np.sin(x)

    usersin.stateless = True

    default_dt = defaultclock.dt
    test_array = np.array([0, 1, 2, 3])
    G = NeuronGroup(
        len(test_array),
        """func = usersin(variable) : 1
                       variable : 1""",
        codeobj_class=NumpyCodeObject,
    )
    G.variable = test_array
    mon = StateMonitor(G, "func", record=True, codeobj_class=NumpyCodeObject)
    net = Network(G, mon)
    net.run(default_dt)

    assert_allclose(np.sin(test_array), mon.func_.flatten())


def test_manual_user_defined_function():
    if prefs.codegen.target != "numpy":
        pytest.skip("numpy-only test")

    default_dt = defaultclock.dt

    # User defined function without any decorators
    def foo(x, y):
        return x + y + 3 * volt

    orig_foo = foo
    # Since the function is not annotated with check units, we need to specify
    # both the units of the arguments and the return unit
    with pytest.raises(ValueError):
        Function(foo, return_unit=volt)
    with pytest.raises(ValueError):
        Function(foo, arg_units=[volt, volt])
    # If the function uses the string syntax for "same units", it needs to
    # specify the names of the arguments
    with pytest.raises(TypeError):
        Function(foo, arg_units=[volt, "x"])
    with pytest.raises(TypeError):
        Function(foo, arg_units=[volt, "x"], arg_names=["x"])  # Needs two entries

    foo = Function(foo, arg_units=[volt, volt], return_unit=volt)

    assert foo(1 * volt, 2 * volt) == 6 * volt

    # a can be any unit, b and c need to be the same unit
    def bar(a, b, c):
        return a * (b + c)

    bar = Function(
        bar,
        arg_units=[None, None, "b"],
        arg_names=["a", "b", "c"],
        return_unit=lambda a, b, c: a * b,
    )
    assert bar(2, 3 * volt, 5 * volt) == 16 * volt
    assert bar(2 * amp, 3 * volt, 5 * volt) == 16 * watt
    assert bar(2 * volt, 3 * amp, 5 * amp) == 16 * watt

    with pytest.raises(DimensionMismatchError):
        bar(2, 3 * volt, 5 * amp)

    # Incorrect argument units
    group = NeuronGroup(
        1,
        """
        dv/dt = foo(x, y)/ms : volt
        x : 1
        y : 1
        """,
    )
    net = Network(group)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms, namespace={"foo": foo})
    assert exc_isinstance(exc, DimensionMismatchError)

    # Incorrect output unit
    group = NeuronGroup(
        1,
        """
        dv/dt = foo(x, y)/ms : 1
        x : volt
        y : volt
        """,
    )
    net = Network(group)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms, namespace={"foo": foo})
    assert exc_isinstance(exc, DimensionMismatchError)

    G = NeuronGroup(
        1,
        """
        func = foo(x, y) : volt
        x : volt
        y : volt
        """,
    )
    G.x = 1 * volt
    G.y = 2 * volt
    mon = StateMonitor(G, "func", record=True)
    net = Network(G, mon)
    net.run(default_dt)

    assert mon[0].func == [6] * volt

    # discard units
    foo.implementations.add_numpy_implementation(orig_foo, discard_units=True)
    G = NeuronGroup(
        1,
        """
        func = foo(x, y) : volt
        x : volt
        y : volt
        """,
    )
    G.x = 1 * volt
    G.y = 2 * volt
    mon = StateMonitor(G, "func", record=True)
    net = Network(G, mon)
    net.run(default_dt)

    assert mon[0].func == [6] * volt


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_manual_user_defined_function_cpp_standalone_compiler_args():
    set_device("cpp_standalone", directory=None)

    @implementation(
        "cpp",
        """
        static inline double foo(const double x, const double y)
        {
            return x + y + _THREE;
        }
        """,  # just check whether we can specify the supported compiler args,
        # only the define macro is actually used
        headers=[],
        sources=[],
        libraries=[],
        include_dirs=[],
        library_dirs=[],
        runtime_library_dirs=[],
        define_macros=[("_THREE", "3")],
    )
    @check_units(x=volt, y=volt, result=volt)
    def foo(x, y):
        return x + y + 3 * volt

    G = NeuronGroup(
        1,
        """
        func = foo(x, y) : volt
        x : volt
        y : volt
        """,
    )
    G.x = 1 * volt
    G.y = 2 * volt
    mon = StateMonitor(G, "func", record=True)
    net = Network(G, mon)
    net.run(defaultclock.dt)
    assert mon[0].func == [6] * volt


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_manual_user_defined_function_cpp_standalone_wrong_compiler_args1():
    set_device("cpp_standalone", directory=None)

    @implementation(
        "cpp",
        """
        static inline double foo(const double x, const double y)
        {
            return x + y + _THREE;
        }
        """,
        some_arg=[],
    )  # non-existing argument
    @check_units(x=volt, y=volt, result=volt)
    def foo(x, y):
        return x + y + 3 * volt

    G = NeuronGroup(
        1,
        """
        func = foo(x, y) : volt
        x : volt
        y : volt
        """,
    )
    mon = StateMonitor(G, "func", record=True)
    net = Network(G, mon)
    with pytest.raises(BrianObjectException) as exc:
        net.run(defaultclock.dt, namespace={"foo": foo})
    assert exc_isinstance(exc, ValueError)


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_manual_user_defined_function_cpp_standalone_wrong_compiler_args2():
    set_device("cpp_standalone", directory=None)

    @implementation(
        "cpp",
        """
        static inline double foo(const double x, const double y)
        {
            return x + y + _THREE;
        }
        """,
        headers="<stdio.h>",
    )  # existing argument, wrong value type
    @check_units(x=volt, y=volt, result=volt)
    def foo(x, y):
        return x + y + 3 * volt

    G = NeuronGroup(
        1,
        """
        func = foo(x, y) : volt
        x : volt
        y : volt
        """,
    )
    mon = StateMonitor(G, "func", record=True)
    net = Network(G, mon)
    with pytest.raises(BrianObjectException) as exc:
        net.run(defaultclock.dt, namespace={"foo": foo})
    assert exc_isinstance(exc, TypeError)


def test_manual_user_defined_function_cython_compiler_args():
    if prefs.codegen.target != "cython":
        pytest.skip("Cython-only test")

    @implementation(
        "cython",
        """
        cdef double foo(double x, const double y):
            return x + y + 3
        """,  # just check whether we can specify the supported compiler args,
        libraries=[],
        include_dirs=[],
        library_dirs=[],
        runtime_library_dirs=[],
    )
    @check_units(x=volt, y=volt, result=volt)
    def foo(x, y):
        return x + y + 3 * volt

    G = NeuronGroup(
        1,
        """
                       func = foo(x, y) : volt
                       x : volt
                       y : volt""",
    )
    G.x = 1 * volt
    G.y = 2 * volt
    mon = StateMonitor(G, "func", record=True)
    net = Network(G, mon)
    net.run(defaultclock.dt)
    assert mon[0].func == [6] * volt


def test_manual_user_defined_function_cython_wrong_compiler_args1():
    if prefs.codegen.target != "cython":
        pytest.skip("Cython-only test")

    @implementation(
        "cython",
        """
        cdef double foo(double x, const double y):
            return x + y + 3
        """,
        some_arg=[],
    )  # non-existing argument
    @check_units(x=volt, y=volt, result=volt)
    def foo(x, y):
        return x + y + 3 * volt

    G = NeuronGroup(
        1,
        """
        func = foo(x, y) : volt
        x : volt
        y : volt
        """,
    )
    mon = StateMonitor(G, "func", record=True)
    net = Network(G, mon)
    with pytest.raises(BrianObjectException) as exc:
        net.run(defaultclock.dt, namespace={"foo": foo})
    assert exc_isinstance(exc, ValueError)


def test_manual_user_defined_function_cython_wrong_compiler_args2():
    if prefs.codegen.target != "cython":
        pytest.skip("Cython-only test")

    @implementation(
        "cython",
        """
        cdef double foo(double x, const double y):
            return x + y + 3
        """,
        libraries="cstdio",
    )  # existing argument, wrong value type
    @check_units(x=volt, y=volt, result=volt)
    def foo(x, y):
        return x + y + 3 * volt

    G = NeuronGroup(
        1,
        """
        func = foo(x, y) : volt
        x : volt
        y : volt
        """,
    )
    mon = StateMonitor(G, "func", record=True)
    net = Network(G, mon)
    with pytest.raises(BrianObjectException) as exc:
        net.run(defaultclock.dt, namespace={"foo": foo})
    assert exc_isinstance(exc, TypeError)


def test_external_function_cython():
    if prefs.codegen.target != "cython":
        pytest.skip("Cython-only test")

    this_dir = os.path.abspath(os.path.dirname(__file__))

    @implementation(
        "cython",
        "from func_def_cython cimport foo",
        sources=[os.path.join(this_dir, "func_def_cython.pyx")],
    )
    @check_units(x=volt, y=volt, result=volt)
    def foo(x, y):
        return x + y + 3 * volt

    G = NeuronGroup(
        1,
        """
        func = foo(x, y) : volt
        x : volt
        y : volt
        """,
    )
    G.x = 1 * volt
    G.y = 2 * volt
    mon = StateMonitor(G, "func", record=True)
    net = Network(G, mon)
    net.run(defaultclock.dt)
    assert mon[0].func == [6] * volt


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_external_function_cpp_standalone():
    set_device("cpp_standalone", directory=None)
    this_dir = os.path.abspath(os.path.dirname(__file__))
    with tempfile.TemporaryDirectory(prefix="brian_testsuite_") as tmpdir:
        # copy the test function to the temporary directory
        # this avoids issues with the file being in a directory that is not writable
        shutil.copy(os.path.join(this_dir, "func_def_cpp.h"), tmpdir)
        shutil.copy(os.path.join(this_dir, "func_def_cpp.cpp"), tmpdir)

        @implementation(
            "cpp",
            "//all code in func_def_cpp.cpp",
            headers=['"func_def_cpp.h"'],
            include_dirs=[tmpdir],
            sources=[os.path.join(tmpdir, "func_def_cpp.cpp")],
        )
        @check_units(x=volt, y=volt, result=volt)
        def foo(x, y):
            return x + y + 3 * volt

        G = NeuronGroup(
            1,
            """
            func = foo(x, y) : volt
            x : volt
            y : volt
            """,
        )
        G.x = 1 * volt
        G.y = 2 * volt
        mon = StateMonitor(G, "func", record=True)
        net = Network(G, mon)
        net.run(defaultclock.dt)
        assert mon[0].func == [6] * volt


@pytest.mark.codegen_independent
def test_user_defined_function_discarding_units():
    # A function with units that should discard units also inside the function
    @implementation("numpy", discard_units=True)
    @check_units(v=volt, result=volt)
    def foo(v):
        return v + 3 * volt  # this normally raises an error for unitless v

    assert foo(5 * volt) == 8 * volt

    # Test the function that is used during a run
    assert foo.implementations[NumpyCodeObject].get_code(None)(5) == 8


@pytest.mark.codegen_independent
def test_user_defined_function_discarding_units_2():
    # Add a numpy implementation explicitly (as in TimedArray)
    unit = volt

    @check_units(v=volt, result=unit)
    def foo(v):
        return v + 3 * unit  # this normally raises an error for unitless v

    foo = Function(pyfunc=foo)

    def unitless_foo(v):
        return v + 3

    foo.implementations.add_implementation("numpy", code=unitless_foo)

    assert foo(5 * volt) == 8 * volt

    # Test the function that is used during a run
    assert foo.implementations[NumpyCodeObject].get_code(None)(5) == 8


@pytest.mark.codegen_independent
def test_function_implementation_container():
    import brian2.codegen.targets as targets

    class ACodeGenerator(CodeGenerator):
        class_name = "A Language"

    class BCodeGenerator(CodeGenerator):
        class_name = "B Language"

    class ACodeObject(CodeObject):
        generator_class = ACodeGenerator
        class_name = "A"

    class A2CodeObject(CodeObject):
        generator_class = ACodeGenerator
        class_name = "A2"

    class BCodeObject(CodeObject):
        generator_class = BCodeGenerator
        class_name = "B"

    # Register the code generation targets
    _previous_codegen_targets = set(targets.codegen_targets)
    targets.codegen_targets = {ACodeObject, BCodeObject}

    @check_units(x=volt, result=volt)
    def foo(x):
        return x

    f = Function(foo)

    container = f.implementations

    # inserting into the container with a CodeGenerator class
    container.add_implementation(BCodeGenerator, code="implementation B language")
    assert container[BCodeGenerator].get_code(None) == "implementation B language"

    # inserting into the container with a CodeObject class
    container.add_implementation(ACodeObject, code="implementation A CodeObject")
    assert container[ACodeObject].get_code(None) == "implementation A CodeObject"

    # inserting into the container with a name of a CodeGenerator
    container.add_implementation("A Language", "implementation A Language")
    assert container["A Language"].get_code(None) == "implementation A Language"
    assert container[ACodeGenerator].get_code(None) == "implementation A Language"
    assert container[A2CodeObject].get_code(None) == "implementation A Language"

    # inserting into the container with a name of a CodeObject
    container.add_implementation("B", "implementation B CodeObject")
    assert container["B"].get_code(None) == "implementation B CodeObject"
    assert container[BCodeObject].get_code(None) == "implementation B CodeObject"

    with pytest.raises(KeyError):
        container["unknown"]

    # some basic dictionary properties
    assert len(container) == 4
    assert {key for key in container} == {
        "A Language",
        "B",
        ACodeObject,
        BCodeGenerator,
    }

    # Restore the previous codegeneration targets
    targets.codegen_targets = _previous_codegen_targets


def test_function_dependencies_cython():
    if prefs.codegen.target != "cython":
        pytest.skip("cython-only test")

    @implementation(
        "cython",
        """
        cdef float foo(float x):
            return 42*0.001
        """,
    )
    @check_units(x=volt, result=volt)
    def foo(x):
        return 42 * mV

    # Second function with an independent implementation for numpy and an
    # implementation for C++ that makes use of the previous function.

    @implementation(
        "cython",
        """
        cdef float bar(float x):
            return 2*foo(x)
        """,
        dependencies={"foo": foo},
    )
    @check_units(x=volt, result=volt)
    def bar(x):
        return 84 * mV

    G = NeuronGroup(5, "v : volt")
    G.run_regularly("v = bar(v)")
    net = Network(G)
    net.run(defaultclock.dt)

    assert_allclose(G.v_[:], 84 * 0.001)


def test_function_dependencies_cython_rename():
    if prefs.codegen.target != "cython":
        pytest.skip("cython-only test")

    @implementation(
        "cython",
        """
        cdef float _foo(float x):
            return 42*0.001
        """,
        name="_foo",
    )
    @check_units(x=volt, result=volt)
    def foo(x):
        return 42 * mV

    # Second function with an independent implementation for numpy and an
    # implementation for C++ that makes use of the previous function.

    @implementation(
        "cython",
        """
        cdef float bar(float x):
            return 2*my_foo(x)
        """,
        dependencies={"my_foo": foo},
    )
    @check_units(x=volt, result=volt)
    def bar(x):
        return 84 * mV

    G = NeuronGroup(5, "v : volt")
    G.run_regularly("v = bar(v)")
    net = Network(G)
    net.run(defaultclock.dt)

    assert_allclose(G.v_[:], 84 * 0.001)


def test_function_dependencies_numpy():
    if prefs.codegen.target != "numpy":
        pytest.skip("numpy-only test")

    @implementation(
        "cpp",
        """
    float foo(float x)
    {
        return 42*0.001;
    }""",
    )
    @check_units(x=volt, result=volt)
    def foo(x):
        return 42 * mV

    # Second function with an independent implementation for C++ and an
    # implementation for numpy that makes use of the previous function.

    # Note that we don't need to use the explicit dependencies mechanism for
    # numpy, since the Python function stores a reference to the referenced
    # function directly

    @implementation(
        "cpp",
        """
        float bar(float x)
        {
            return 84*0.001;
        }""",
    )
    @check_units(x=volt, result=volt)
    def bar(x):
        return 2 * foo(x)

    G = NeuronGroup(5, "v : volt")
    G.run_regularly("v = bar(v)")
    net = Network(G)
    net.run(defaultclock.dt)

    assert_allclose(G.v_[:], 84 * 0.001)


@pytest.mark.standalone_compatible
def test_repeated_function_dependencies():
    # each of the binomial functions adds randn as a depency, see #988
    test_neuron = NeuronGroup(
        1,
        "x : 1",
        namespace={
            "bino_1": BinomialFunction(10, 0.5),
            "bino_2": BinomialFunction(10, 0.6),
        },
    )
    test_neuron.x = "bino_1()+bino_2()"

    run(0 * ms)


@pytest.mark.standalone_compatible
def test_binomial():
    binomial_f_approximated = BinomialFunction(100, 0.1, approximate=True)
    binomial_f = BinomialFunction(100, 0.1, approximate=False)

    # Just check that it does not raise an error and that it produces some
    # values
    G = NeuronGroup(
        1,
        """
        x : 1
        y : 1
        """,
    )
    G.run_regularly(
        """
        x = binomial_f_approximated()
        y = binomial_f()
        """
    )
    mon = StateMonitor(G, ["x", "y"], record=0)
    run(1 * ms)
    assert np.var(mon[0].x) > 0
    assert np.var(mon[0].y) > 0


@pytest.mark.standalone_compatible
def test_poisson():
    # Just check that it does not raise an error and that it produces some
    # values
    G = NeuronGroup(
        5,
        """
        l : 1
        x : integer
        y : integer
        z : integer
        """,
    )
    G.l = [0, 1, 5, 15, 25]
    G.run_regularly(
        """
        x = poisson(l)
        y = poisson(5)
        z = poisson(0)
        """
    )
    mon = StateMonitor(G, ["x", "y", "z"], record=True)
    run(100 * defaultclock.dt)
    assert_equal(mon.x[0], 0)
    assert all(np.var(mon.x[1:], axis=1) > 0)
    assert all(np.var(mon.y, axis=1) > 0)
    assert_equal(mon.z, 0)


def test_declare_types():
    if prefs.codegen.target != "numpy":
        pytest.skip("numpy-only test")

    @declare_types(a="integer", b="float", result="highest")
    def f(a, b):
        return a * b

    assert f._arg_types == ["integer", "float"]
    assert f._return_type == "highest"

    @declare_types(b="float")
    def f(a, b, c):
        return a * b * c

    assert f._arg_types == ["any", "float", "any"]
    assert f._return_type == "float"

    def bad_argtype():
        @declare_types(b="floating")
        def f(a, b, c):
            return a * b * c

    with pytest.raises(ValueError):
        bad_argtype()

    def bad_argname():
        @declare_types(d="floating")
        def f(a, b, c):
            return a * b * c

    with pytest.raises(ValueError):
        bad_argname()

    @check_units(a=volt, b=1)
    @declare_types(a="float", b="integer")
    def f(a, b):
        return a * b

    @declare_types(a="float", b="integer")
    @check_units(a=volt, b=1)
    def f(a, b):
        return a * b

    def bad_units():
        @declare_types(a="integer", b="float")
        @check_units(a=volt, b=1, result=volt)
        def f(a, b):
            return a * b

        eqs = """
        dv/dt = f(v, 1)/second : 1
        """
        G = NeuronGroup(1, eqs)
        Network(G).run(1 * ms)

    with pytest.raises(BrianObjectException) as exc:
        bad_units()
    assert exc_isinstance(exc, TypeError)

    def bad_type():
        @implementation("numpy", discard_units=True)
        @declare_types(a="float", result="float")
        @check_units(a=1, result=1)
        def f(a):
            return a

        eqs = """
        a : integer
        dv/dt = f(a)*v/second : 1
        """
        G = NeuronGroup(1, eqs)
        Network(G).run(1 * ms)

    with pytest.raises(BrianObjectException) as exc:
        bad_type()
    assert exc_isinstance(exc, TypeError)


def test_multiple_stateless_function_calls():
    # Check that expressions such as rand() + rand() (which might be incorrectly
    # simplified to 2*rand()) raise an error
    G = NeuronGroup(1, "dv/dt = (rand() - rand())/second : 1")
    net = Network(G)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms)
    assert exc_isinstance(exc, NotImplementedError)
    G2 = NeuronGroup(1, "v:1", threshold="v>1", reset="v=rand() - rand()")
    net2 = Network(G2)
    with pytest.raises(BrianObjectException) as exc:
        net2.run(0 * ms)
    assert exc_isinstance(exc, NotImplementedError)
    G3 = NeuronGroup(1, "v:1")
    G3.run_regularly("v = rand() - rand()")
    net3 = Network(G3)
    with pytest.raises(BrianObjectException) as exc:
        net3.run(0 * ms)
    assert exc_isinstance(exc, NotImplementedError)
    G4 = NeuronGroup(1, "x : 1")
    # Verify that synaptic equations are checked as well, see #1146
    S = Synapses(G4, G4, "dy/dt = (rand() - rand())/second : 1 (clock-driven)")
    S.connect()
    net = Network(G4, S)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms)
    assert exc_isinstance(exc, NotImplementedError)


@pytest.mark.codegen_independent
def test_parse_dimension_errors():
    from brian2.parsing.expressions import parse_expression_dimensions

    @check_units(x=1, result=1)
    def foo(x):
        return x

    # Function call with keyword arguments
    with pytest.raises(ValueError):
        parse_expression_dimensions("foo(a=1, b=2)", {"foo": foo})
    # Unknown function
    with pytest.raises(SyntaxError):
        parse_expression_dimensions("bar(1, 2)", {"foo": foo})
    # Function without unit definition
    with pytest.raises(ValueError):
        parse_expression_dimensions("bar(1, 2)", {"bar": lambda x, y: x + y})
    # Function with wrong number of arguments
    with pytest.raises(SyntaxError):
        parse_expression_dimensions("foo(1, 2)", {"foo": foo})


if __name__ == "__main__":
    # prefs.codegen.target = 'numpy'
    import time

    from _pytest.outcomes import Skipped

    from brian2 import prefs

    for f in [
        test_constants_sympy,
        test_constants_values,
        test_math_functions,
        test_clip,
        test_bool_to_int,
        test_timestep_function,
        test_timestep_function_during_run,
        test_user_defined_function,
        test_user_defined_function_units,
        test_simple_user_defined_function,
        test_manual_user_defined_function,
        test_external_function_cython,
        test_user_defined_function_discarding_units,
        test_user_defined_function_discarding_units_2,
        test_function_implementation_container,
        test_function_dependencies_numpy,
        test_function_dependencies_cython,
        test_function_dependencies_cython_rename,
        test_repeated_function_dependencies,
        test_binomial,
        test_poisson,
        test_declare_types,
        test_multiple_stateless_function_calls,
    ]:
        try:
            start = time.time()
            f()
            print("Test", f.__name__, "took", time.time() - start)
        except Skipped as e:
            print("Skipping test", f.__name__, e)



import re
import os
import pkg_resources
import tomllib
import json
import sys


if __name__ == "__main__":
    if not len(sys.argv) == 2:
        raise ValueError("Usage: python create_codemeta.py <version>")
    basedir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")

    with open(os.path.join(basedir, "pyproject.toml"), "rb") as f:
        pyproject = tomllib.load(f)
    with open(os.path.join(basedir, ".codemeta", "codemeta_base.json"), "r", encoding="utf-8") as f:
        codemeta = json.load(f)
    with open(os.path.join(basedir, "CONTRIBUTORS"), "r", encoding="utf-8") as f:
        contributors = f.read().splitlines()[4:]  # Skip comments at the beginning

    # Add software requirements from pyproject.toml
    parsed_deps = pkg_resources.parse_requirements(pyproject['project']['dependencies'])
    codemeta["softwareRequirements"] = []
    for dep in parsed_deps:
        version = ",".join(f"{op}{v}" for op,v in dep.specs)
        requirement = {"name": dep.project_name,"@type": "SoftwareApplication", "runtimePlatform": "Python 3"}
        if version:
            requirement["version"] = version
        codemeta["softwareRequirements"].append(requirement)

    # Add contributors from AUTHORS
    codemeta["contributor"] = []
    for contributor in contributors:
        matches = re.match(r"^(\w[\w-]*?) ([\w]+[.]? )??(\w+) \(@(.*)\)$", contributor)
        if not matches:
            raise ValueError("author not matched:", contributor)
        given_name, middle_name, family_name, github = matches.groups()

        contributor = {"@type": "Person", "givenName": given_name, "familyName": family_name, "identifier": f"https://github.com/{github}"}
        # FIXME: additionalName does not seem to be recognized by codemeta (validation fails)
        if middle_name:
            contributor["givenName"] += " " + middle_name.strip()
        codemeta["contributor"].append(contributor)

    # Add version from setuptools_scm
    version = sys.argv[1]
    codemeta["version"] = version
    codemeta["softwareVersion"] = version

    # Write codemeta.json
    with open(os.path.join(basedir, "codemeta.json"), "w", encoding="utf-8") as f:
        json.dump(codemeta, f, indent=2, ensure_ascii=False)



import pytest

from brian2 import *


def test_error_message():
    if prefs.codegen.target != "numpy":
        pytest.skip("numpy-only test")

    @check_units(x=1, result=1)
    def foo(x):
        raise ValueError()

    G = NeuronGroup(1, "v : 1")
    G.run_regularly("v = foo(3)")
    with pytest.raises(BrianObjectException) as exc:
        run(defaultclock.dt)
        # The actual code line should be mentioned in the error message
        exc.match("v = foo(3)")


if __name__ == "__main__":
    prefs.codegen.target = "numpy"
    test_error_message()
